package coco;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectDescription;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.Path;
import org.eclipse.equinox.app.IApplication;
import org.eclipse.equinox.app.IApplicationContext;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.JavaModelException;

import coco.listener.CoCoListener;
import coco.log.Log;
import coco.task.EmptyProgressMonitor;
import coco.task.ITask;
import coco.task.etc.CheckIfValidCUASTV;
import coco.task.etc.RemoveCommentsTask;
import coco.task.etc.RemoveExistingInstrumentationTask;
import coco.task.extract.pct.ExtractPredicatesTask;
import coco.task.instantiate.InsertBasicBlocksTask;
import coco.task.instantiate.InstantiateTask;
import coco.task.renamevars.RenameVarsTask;
import coco.util.ASTVHelper;
import coco.util.CoCoUtil;
import coring.OptParser;
import coring.OptParserStatus;

/**
 * Entry point to instrumentation.
 * 
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 * @author Rohan Sharma (sharma.rohan1990@gmail.com)
 */
public class CoCoInstrumentation implements IApplication {

    private IProject project;
    private CoCoInstrumentationOpt options;
    private CoCoListener listener = new CoCoListener();

    @Override
    public Object start(IApplicationContext context) {
        listener.bcExecutionStarted();
        try {
            String[] args = (String[]) context.getArguments().get("application.args");
            options = new CoCoInstrumentationOpt();
            OptParserStatus status = OptParser.helpOrParse(options, args);
            if (status.isSuccessfullyParsed()) {
                Log.open(options.logFileName.g());
                if ((options.isExtractPredicates.g() || options.isRenameVariables.g()) && options.isCopyOrigProject()) {
                    copyOriginalProject();
                }
                openProject();
                refreshProject();
                listener.bcInstrumentationStarted(options, project);
                instrument();
                closeProject();
            } else {
                System.out.println(status.getExceptionMsg());
            }
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        listener.bcExecutionFinished();
        Log.close();
        return null;
    }

    @Override
    public void stop() {
        // NOP
    }

    private void copyOriginalProject() throws CoreException {
        IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
        IProject originalProject = root.getProject(options.projectName.g());
        if (!originalProject.exists()) {
            IProjectDescription description = ResourcesPlugin.getWorkspace().loadProjectDescription(
                    new Path(options.projectLocation.g()));
            originalProject.create(description, IProject.FORCE, EmptyProgressMonitor.inst());
            // throw new RuntimeException("Project does not exist!");
        }
        originalProject.open(EmptyProgressMonitor.inst());
        originalProject.copy(new Path(options.newProjectName.g()), true, EmptyProgressMonitor.inst());
        originalProject.delete(false, true, EmptyProgressMonitor.inst());
        // originalProject.close(CoverageProgressMonitor.inst());
    }

    private void openProject() throws CoreException {
        IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
        project = root.getProject(options.newProjectName.g());
        if (!project.exists()) {
            IProjectDescription description = project.getWorkspace().loadProjectDescription(
                    new Path(options.projectLocation.g()));
            project.create(description, EmptyProgressMonitor.inst());
        }
        project.open(EmptyProgressMonitor.inst());
    }

    private void refreshProject() throws CoreException {
        project.refreshLocal(IResource.DEPTH_INFINITE, EmptyProgressMonitor.inst());
    }

    private void closeProject() throws CoreException {
        project.delete(false, true, EmptyProgressMonitor.inst());
        // project.close(new NullProgressMonitor());
    }

    private void instrument() throws Exception {
        List<ICompilationUnit> icus = CoCoUtil.findCompilationUnitsInProject(project);
        filterICompilationUnitsNotToInstrument(icus);
        if (icus.size() == 0) {
            listener.bcNoClassNameIsSpecified();
            return;
        }

        List<ITask> tasks = createTasks();
        if (tasks.size() == 0) {
            listener.bcNoTaskIsSpecified();
            return;
        }

        CheckIfValidCUASTV checkIfValidCU = new CheckIfValidCUASTV();
        for (ICompilationUnit icu : icus) {
            listener.bcInstrumentationOfICUStarted(icu);
            ASTVHelper.acceptVisitor(project, icu, checkIfValidCU, false);
            if (checkIfValidCU.isValid()) {
                for (ITask task : tasks) {
                    listener.bcTaskStarted(task);
                    task.perform(icu);
                    listener.bcTaskFinished(task);
                }
            }
            listener.bcInstrumentationOfICUFinished();
        }
    }

    /**
     * Create tasks as specified by the options.
     */
    private List<ITask> createTasks() {
        List<ITask> tasks = new LinkedList<ITask>();
        if (options.isRemoveComments.g()) {
            tasks.add(new RemoveCommentsTask(project));
        }
        if (options.isRemoveExistingInstrumentation() || options.isPCTMS.g()) {
            tasks.add(new RemoveExistingInstrumentationTask(project));
        }
        if (options.isRenameVariables.g()) {
            tasks.add(new RenameVarsTask(project));
        }
        if (options.isExtractPredicates.g() && options.isPCT()) {
            tasks.add(new ExtractPredicatesTask(options, project));
        }
        if (options.isInsertBasicBlocks.g()) {
            tasks.add(new InsertBasicBlocksTask(options, project));
        }
        if (options.isInstantiatePredicates.g()) {
            tasks.add(new InstantiateTask(options, project));
        }
        return tasks;
    }

    private void filterICompilationUnitsNotToInstrument(List<ICompilationUnit> icus) throws JavaModelException {
        Iterator<ICompilationUnit> iterator = icus.iterator();
        while (iterator.hasNext()) {
            ICompilationUnit icu = iterator.next();
            String fullyQualifiedName = "";
            if (icu.getPackageDeclarations().length > 0) {
                fullyQualifiedName = icu.getPackageDeclarations()[0].getElementName() + ".";
            }
            fullyQualifiedName += icu.getElementName().replace(".java", "");
            if (!options.extractFrom().contains(fullyQualifiedName)) {
                iterator.remove();
            }
        }
    }

}
