package coco;

import java.util.List;

import coring.BooleanOpt;
import coring.IntegerOpt;
import coring.OptHelp;
import coring.OptRequired;
import coring.StringLst;
import coring.StringOpt;

/**
 * Description of command line options.
 * 
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 * @author Rohan Sharma (sharma.rohan1990@gmail.com)
 */
public class CoCoInstrumentationOpt {

    // CONSTANTS

    public static final String COCO_RUNTIME_PCK = "coco.runtime";
    public static final String COCO_RUNTIME_CLASS = "CoCoRuntime";

    public static final String INST_METHOD = "inst";
    public static final String TOUCH_METHOD = "touch";
    public static final String TOUCH_BRANCH_METHOD = "touchBranch";
    public static final String TOUCH_START_METHOD = "touchStartMethod";

    public static final String TOUCH_BEGIN_METHOD = "touchBeginMethod";
    public static final String TOUCH_END_METHOD = "touchEndMethod";
    public static final String TOUCH_BEGIN_STATIC = "touchBeginStatic";
    public static final String TOUCH_END_STATIC = "touchEndStatic";

    public static final String SET_IN_PREDICATE_METHOD = "testAndSetInPredicate";
    public static final String RESET_IN_PREDICATE_METHOD = "resetInPredicate";
    public static final String IGNORE_TOUCH_METHOD = "ignoreNextTouch";

    // PROJECTS INFO

    @OptHelp("List of fully qualified names to be instrumented.")
    @OptRequired
    public final StringLst compilationUnits = new StringLst("compilation-units");

    @OptHelp("Name of the project that contains compilation units to be instrumented.")
    @OptRequired
    public final StringOpt projectName = new StringOpt("project-name");

    @OptHelp("Location of the project that contains compilation units to be instrumented.")
    @OptRequired
    public final StringOpt projectLocation = new StringOpt("project-location");

    @OptHelp("Name of the project that will contain instrumented compilation units.")
    public final StringOpt newProjectName = new StringOpt("new-project-name");

    // MAKE INSTRUMENTATION OF A PROJECT POSSIBLE

    @OptHelp("Instrument project in place (needed if not all files can be instrumented at once, e.g., OOM).")
    public final BooleanOpt isDoNotCopyProject = new BooleanOpt("do-not-copy-project").s(false);

    @OptHelp("If true counter is not reset for each icu (should be used multiple files from a project are instrumented).")
    public final BooleanOpt isContinuousPredicateCounting = new BooleanOpt("continuous-predicate-counting").s(false);

    @OptHelp("What is the starting id of the predicate for this run (when multiple files from a project are instrumented).")
    public final IntegerOpt basePredicateId = new IntegerOpt("base-predicate-id").s(1);

    // TASKS

    @OptHelp("Include task that removes comments.")
    public final BooleanOpt isRemoveComments = new BooleanOpt("remove-comments").s(false);

    @OptHelp("Include task that removes existing instrumentation and leaves inst() calls.")
    public final BooleanOpt isRemoveExistingInstrumentation = new BooleanOpt("remove-existing-instrumentation")
            .s(false);

    @OptHelp("Include task that renames variables.")
    public final BooleanOpt isRenameVariables = new BooleanOpt("rename-variables").s(true);

    @OptHelp("Include task that extracts predicates.")
    public final BooleanOpt isExtractPredicates = new BooleanOpt("extract-predicates").s(false);

    @OptHelp("Include task that inserts basic blocks at appropriate points.")
    public final BooleanOpt isInsertBasicBlocks = new BooleanOpt("insert-basic-blocks").s(false);

    @OptHelp("Include task that instantiates predicates.")
    public final BooleanOpt isInstantiatePredicates = new BooleanOpt("instantiate-predicates").s(false);

    // INSTANTIATION

    @OptHelp("Instantiate predicates at each statement.")
    public final BooleanOpt isPCTST = new BooleanOpt("instantiate-statement").s(false);

    @OptHelp("Branch coverage instrumentation.")
    public final BooleanOpt isBranchCoverage = new BooleanOpt("branch-coverage").s(false);

    @OptHelp("Statement coverage instrumentation.")
    public final BooleanOpt isStatementCoverage = new BooleanOpt("statement-coverage").s(false);

    @OptHelp("Instantiate predicates at points as it was done in FASE 2011.")
    public final BooleanOpt isPCTMS = new BooleanOpt("instantiate-FASE").s(false);

    @OptHelp("Instantiate atomic predicates.")
    public final BooleanOpt isAtomicPredicates = new BooleanOpt("atomic-predicates").s(true);

    // LOGS

    @OptHelp("Name of the file where the results should be saved.")
    public final StringOpt fileWithInstrumentationDetails = new StringOpt("instrumentation-details")
            .s(".coverage-instrumentation-info");

    @OptHelp("Execution/trace log file.")
    public final StringOpt logFileName = new StringOpt("log-file-name").s(".coverage-instrumentation-log");

    // OUTPUT

    @OptHelp("Print results as latex macros.")
    public final BooleanOpt isUseNewcommand = new BooleanOpt("use-newcommand").s(false);

    @OptHelp("String to prefix macro for newcommand.")
    public final StringOpt newcommandPrefix = new StringOpt("newcommand-prefix").s("");

    @OptHelp("String to suffix macro for newcommand.")
    public final StringOpt newcommandSuffix = new StringOpt("newcommand-suffix").s("");

    @OptHelp("Output more info during instrumentation.")
    public final BooleanOpt isVerbose = new BooleanOpt("verbose");

    // ACCESS TO OPTIONS

    public List<String> extractFrom() {
        return compilationUnits.g();
    }

    public boolean isCopyOrigProject() {
        return !isDoNotCopyProject.g();
    }

    public boolean isInstantiateBasicBlock() {
        return !isPCTST.g() && !isPCTMS.g();
    }

    public String getFileWithInstrumentationDetails() {
        return fileWithInstrumentationDetails.g();
    }

    public int getBasePredicateId() {
        return basePredicateId.g();
    }

    public boolean isContinuousPredicateCounting() {
        return isContinuousPredicateCounting.g();
    }

    public boolean isRemoveExistingInstrumentation() {
        return isRemoveExistingInstrumentation.g();
    }

    public boolean isPCT() {
        return !(isBranchCoverage.g() || isStatementCoverage.g());
    }
}
