package coco.listener;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.PrintWriter;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.IPath;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IPackageDeclaration;
import org.eclipse.jdt.core.JavaModelException;

import coco.CoCoInstrumentationOpt;
import coco.task.ITask;
import coco.task.etc.RemoveCommentsTask;
import coco.task.extract.pct.ExtractPredicatesTask;
import coco.task.instantiate.InsertBasicBlocksTask;
import coco.task.instantiate.InstantiateTask;
import coco.task.renamevars.RenameVarsTask;
import coring.Newcommand;

/**
 * Notified for important events. Subclasses can extend behavior.
 * 
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 * @author Rohan Sharma (sharma.rohan1990@gmail.com)
 */
public class CoCoListener {

    /** Separator in file that keeps instrumentation details */
    private static final String SEPARATOR = " ";

    /** Header of the message at the end of instrumentation */
    private static final String HEADER_SEPARATOR = "\n//////////////////////////////";

    /** Sign between key and value */
    private static final String SIGN_BTW = "=";

    /** Default indentation */
    private static final String INDENT = "  ";

    /** Text to be written with the results of the instrumentation */
    private static final String NUM_OF_POINTS = "NumOfPoints";
    private static final String NUM_OF_PREDICATES = "NumOfPredicates";
    private static final String TOTAL_NUM_OF_POINTS = "TotalNumOfPoints";
    private static final String TOTAL_NUM_OF_PREDICATES = "TotalNumOfPredicates";
    private static final String INSTRUMENTATION_TIME = "InstrumentationTime";

    /** Variables to save info between multiple runs (needed for large projects) */
    private static final String NUM_OF_EXTRACTED_PREDICATES = "NumOfExtracted";
    private static final String NEXT_PREDICATE = "NextPredicate";
    private static final String CONTINUOUS_NUM_OF_POINTS = "ContinuousNumOfPoints";

    /** Application arguments */
    private CoCoInstrumentationOpt options;

    /** Time at the beginning of execution */
    private long startTime;

    /** Time at the beginning of the current task */
    private long startTaskTime;

    /** Number of points in the current icu */
    private long numOfPoints;

    /** Total number of points created during current instrumentation */
    private int totalNumOfPoints;

    /** Number of predicates in the current icu */
    private int numOfPredicates;

    /** Total number of predicates created during current instrumentation */
    private int totalNumOfPredicates;

    /** Number of extracted predicates (i.e., total number in all runs) */
    private int numOfExtractedPredicates;

    /** Number of points if continuous counting is used */
    private long continuousNumOfPoints;

    /** Connection to the file with details about instrumentation */
    private PrintWriter pw;

    /** Currently analyzed ICU */
    private ICompilationUnit icu;

    public void bcExecutionStarted() {
        startTime = System.currentTimeMillis();
    }

    public void bcExecutionFinished() {
        long executionTime = System.currentTimeMillis() - startTime;
        println(HEADER_SEPARATOR);
        if (options.isUseNewcommand.g()) {
            println(Newcommand.makeNewcommand(prefixMacroSuffix(TOTAL_NUM_OF_POINTS), totalNumOfPoints));
        } else {
            String s = String.format("%s %s %d", prefixMacroSuffix(TOTAL_NUM_OF_POINTS), SIGN_BTW, totalNumOfPoints);
            println(s);
        }
        if (options.isUseNewcommand.g()) {
            println(Newcommand.makeNewcommand(prefixMacroSuffix(TOTAL_NUM_OF_PREDICATES), totalNumOfPredicates));
        } else {
            String s = String.format("%s %s %d", prefixMacroSuffix(TOTAL_NUM_OF_PREDICATES), SIGN_BTW,
                    totalNumOfPredicates);
            println(s);
        }
        if (options.isUseNewcommand.g()) {
            println(Newcommand.makeNewcommand(prefixMacroSuffix(INSTRUMENTATION_TIME), executionTime));
        } else {
            String s = String.format("%s %s %d", prefixMacroSuffix(INSTRUMENTATION_TIME), SIGN_BTW, executionTime);
            println(s);
        }
        if (options.isUseNewcommand.g()) {
            println(Newcommand.makeNewcommand(prefixMacroSuffix(NUM_OF_EXTRACTED_PREDICATES), numOfExtractedPredicates));
        } else {
            String s = String.format("%s %s %d", prefixMacroSuffix(NUM_OF_EXTRACTED_PREDICATES), SIGN_BTW,
                    numOfExtractedPredicates);
            println(s);
        }
        println("");
        if (pw != null) {
            pw.println(NEXT_PREDICATE + SEPARATOR + (numOfExtractedPredicates + 1));
            pw.println(CONTINUOUS_NUM_OF_POINTS + SEPARATOR + continuousNumOfPoints);
            pw.flush();
            pw.close();
        }
    }

    public void bcInstrumentationStarted(CoCoInstrumentationOpt args, IProject project) {
        this.options = args;
        try {
            IPath path = project.getFile(args.getFileWithInstrumentationDetails()).getLocation();
            File fileWithInstrumentationDetails = path.toFile();
            // reload number of points and id for the next predicate
            if (fileWithInstrumentationDetails.exists()) {
                BufferedReader br = new BufferedReader(new FileReader(fileWithInstrumentationDetails));
                String s;
                while ((s = br.readLine()) != null) {
                    if (s.contains(CONTINUOUS_NUM_OF_POINTS)) {
                        String[] parts = s.split(SEPARATOR);
                        continuousNumOfPoints = Long.parseLong(parts[parts.length - 1]);
                    }
                    if (s.contains(NEXT_PREDICATE)) {
                        String[] parts = s.split(SEPARATOR);
                        numOfExtractedPredicates = Integer.parseInt(parts[parts.length - 1]) - 1;
                    }
                }
            }
            FileOutputStream fos = new FileOutputStream(fileWithInstrumentationDetails, true);
            pw = new PrintWriter(fos);
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    public void bcInstrumentationOfICUStarted(ICompilationUnit icu) {
        this.icu = icu;
        try {
            IPackageDeclaration[] packageDeclarations = icu.getPackageDeclarations();
            String packageName = packageDeclarations.length > 0 ? packageDeclarations[0].getElementName() + "." : "";
            printlnIfVerbose("Started Instrumenting " + packageName + icu.getElementName().replaceAll("\\.java", ""));
            if (!options.isVerbose.g()) {
                print(".");
            }
        } catch (JavaModelException e) {
            e.printStackTrace();
        }
    }

    public void bcInstrumentationOfICUFinished() {
        // TODO: we may want to include just the name of the class
        String className = icu.getElementName().replaceAll("\\.java", "").replaceAll("FASE", "");
        if (options.isUseNewcommand.g()) {
            printlnIfVerbose(Newcommand.makeNewcommand(macroSuffix(className + NUM_OF_POINTS), numOfPoints));
        } else {
            String s = String.format("%s %s %d", macroSuffix(className + NUM_OF_POINTS), SIGN_BTW, numOfPoints);
            printlnIfVerbose(s);
        }
        if (options.isUseNewcommand.g()) {
            printlnIfVerbose(Newcommand.makeNewcommand(macroSuffix(className + NUM_OF_PREDICATES), numOfPredicates));
        } else {
            String s = String.format("%s %s %d", macroSuffix(className + NUM_OF_PREDICATES), SIGN_BTW, numOfPredicates);
            printlnIfVerbose(s);
        }
        if (pw != null) {
            pw.println(icu.getPath() + SEPARATOR + NUM_OF_POINTS + SEPARATOR + numOfPoints);
            pw.println(icu.getPath() + SEPARATOR + NUM_OF_PREDICATES + SEPARATOR + numOfPredicates);
        }
        numOfPredicates = 0;
        numOfPoints = 0;
    }

    public void bcTaskStarted(ITask task) {
        // invoke appropriate method based on the task type
        if (task instanceof RemoveCommentsTask) {
            bcTaskStarted((RemoveCommentsTask) task);
        }
        if (task instanceof RenameVarsTask) {
            bcTaskStarted((RenameVarsTask) task);
        }
        if (task instanceof ExtractPredicatesTask) {
            bcTaskStarted((ExtractPredicatesTask) task);
        }
        if (task instanceof InsertBasicBlocksTask) {
            bcTaskStarted((InsertBasicBlocksTask) task);
        }
        if (task instanceof InstantiateTask) {
            bcTaskStarted((InstantiateTask) task);
        }
        // start measuring time for the task
        printlnIfVerbose(INDENT + task.getClass().getName() + "...");
        startStopwatch(task);
    }

    public void bcTaskFinished(ITask task) {
        // stop measuring time for the task
        printlnIfVerbose(INDENT + INDENT + (System.currentTimeMillis() - startTaskTime) + " [ms]");
        // invoke appropriate method based on the task type
        if (task instanceof RemoveCommentsTask) {
            bcTaskFinished((RemoveCommentsTask) task);
        }
        if (task instanceof RenameVarsTask) {
            bcTaskFinished((RenameVarsTask) task);
        }
        if (task instanceof ExtractPredicatesTask) {
            bcTaskFinished((ExtractPredicatesTask) task);
        }
        if (task instanceof InsertBasicBlocksTask) {
            bcTaskFinished((InsertBasicBlocksTask) task);
        }
        if (task instanceof InstantiateTask) {
            bcTaskFinished((InstantiateTask) task);
        }
    }

    protected void bcTaskStarted(RemoveCommentsTask task) {
        // see subclasses
    }

    protected void bcTaskFinished(RemoveCommentsTask task) {
        // see subclasses
    }

    protected void bcTaskStarted(RenameVarsTask task) {
        // see subclasses
    }

    protected void bcTaskFinished(RenameVarsTask task) {
        // see subclasses
    }

    protected void bcTaskStarted(ExtractPredicatesTask task) {
        // see subclasses
    }

    protected void bcTaskFinished(ExtractPredicatesTask task) {
        // there is no need for += here; see ExtractedPredicatesTask
        numOfExtractedPredicates = task.getNumOfExtractedPredicates();
    }

    protected void bcTaskStarted(InsertBasicBlocksTask task) {
        // see subclasses
    }

    protected void bcTaskFinished(InsertBasicBlocksTask task) {
        // see subclasses
    }

    protected void bcTaskStarted(InstantiateTask task) {
        task.setContinuousNumOfPoints(continuousNumOfPoints);
    }

    protected void bcTaskFinished(InstantiateTask task) {
        numOfPoints = task.getNumOfPoints();
        numOfPredicates = task.getNumOfPredicates();
        totalNumOfPoints += task.getNumOfPoints();
        totalNumOfPredicates += task.getNumOfPredicates();
        continuousNumOfPoints = task.getContinuousNumOfPoints();
    }

    public void bcNoClassNameIsSpecified() {
        println("WARNING: no class is specified (use fully qualified names)");
    }

    public void bcNoTaskIsSpecified() {
        println("WARNING: no task is specified.");
    }

    private String macroSuffix(String macro) {
        return macro + options.newcommandSuffix.g();
    }

    private String prefixMacroSuffix(String macro) {
        return options.newcommandPrefix.g() + macro + options.newcommandSuffix.g();
    }

    private void startStopwatch(ITask task) {
        startTaskTime = System.currentTimeMillis();
    }

    private void printlnIfVerbose(String s) {
        if (options.isVerbose.g()) {
            System.out.println(s);
        }
    }

    private void print(String s) {
        System.out.print(s);
    }

    private void println(String s) {
        System.out.println(s);
    }
}
