package coco.task.etc;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.AnonymousClassDeclaration;
import org.eclipse.jdt.core.dom.ArrayAccess;
import org.eclipse.jdt.core.dom.Block;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.FieldAccess;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.IfStatement;
import org.eclipse.jdt.core.dom.InfixExpression;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.Name;
import org.eclipse.jdt.core.dom.QualifiedName;
import org.eclipse.jdt.core.dom.ReturnStatement;
import org.eclipse.jdt.core.dom.Statement;
import org.eclipse.jdt.core.dom.TryStatement;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.jdt.internal.core.ResolvedBinaryField;
import org.eclipse.jdt.internal.core.ResolvedBinaryType;
import org.eclipse.jdt.internal.core.ResolvedSourceField;
import org.eclipse.jdt.internal.core.ResolvedSourceType;

import coco.CoCoASTV;
import coco.task.extract.pct.ConstructPredicatesASTV;
import coco.util.ASTBuilder;
import coco.util.CoCoASTBuilder;
import coco.util.CoCoASTMatcher;
import coco.util.CoCoUtil;

/**
 * Visitor that inserts if statements to check for null dereference, index out
 * of bounds, and recursive predicate calls. The visitor is invoked after
 * predicates are extracted. Checks are created for all appropriate expressions.
 * (Duplicates are not removed/detected at this point.)
 * 
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 * @author Rohan Sharma (sharma.rohan1990@gmail.com)
 */
@SuppressWarnings("restriction")
public class AddNullIOBChecksASTV extends CoCoASTV {

    private static final String LENGHT_TOKEN = "length";
    private static final String ZERO_TOKEN = "0";
    private static final String MINUS_TOKEN = "-";

    /** True if visiting predicate */
    private boolean visitingPredicate;

    /** New list of statements for a predicate */
    private List<Statement> statements;

    /** Id of the current predicate */
    private String predicateId;

    /** Mapping from type sequence number in CU to conditions defined in it */
    private Map<Integer, List<Expression>> typeSeqToConditions;

    /** Mapping from condition to id of that condition */
    private Map<Expression, String> conditionToId;

    /** Current types on stack */
    private final Stack<Integer> stackOfTypes;

    /** Conditions of the current type */
    private List<Expression> conditions;

    /** Next type in sequence */
    private int nextTypeInSeq = 0;

    public AddNullIOBChecksASTV(ConstructPredicatesASTV constructPredicatesV) {
        this.typeSeqToConditions = constructPredicatesV.getTypeSeqToConditions();
        this.conditionToId = constructPredicatesV.getConditionToId();
        this.stackOfTypes = new Stack<Integer>();
    }

    // ASTVisitor interface

    @Override
    public final boolean visit(AnonymousClassDeclaration node) {
        nextTypeInSeq++;
        stackOfTypes.push(nextTypeInSeq);
        conditions = typeSeqToConditions.get(nextTypeInSeq);
        return true;
    }

    @Override
    public final void endVisit(AnonymousClassDeclaration node) {
        stackOfTypes.pop();
        if (!stackOfTypes.isEmpty()) {
            conditions = typeSeqToConditions.get(stackOfTypes.peek());
        }
    }

    @Override
    public final boolean oVisit(TypeDeclaration node) {
        nextTypeInSeq++;
        stackOfTypes.push(nextTypeInSeq);
        conditions = typeSeqToConditions.get(nextTypeInSeq);
        return true;
    }

    @Override
    public final void oEndVisit(TypeDeclaration node) {
        stackOfTypes.pop();
        if (!stackOfTypes.isEmpty()) {
            conditions = typeSeqToConditions.get(stackOfTypes.peek());
        }
    }

    @Override
    public boolean visit(QualifiedName node) {
        if (visitingPredicate && !isAccessUsingClass(node.getQualifier())) {
            IJavaElement javaElement = node.resolveBinding().getJavaElement();
            if (javaElement instanceof ResolvedSourceField || javaElement instanceof ResolvedBinaryField) {
                createIfNullReturnPredicateId(node.getQualifier());
            }
        }
        return true;
    }

    @Override
    public boolean visit(FieldAccess node) {
        if (visitingPredicate && !isAccessUsingClass(node.getExpression())) {
            createIfNullReturnPredicateId(node.getExpression());
        }
        return true;
    }

    @Override
    public boolean visit(MethodInvocation node) {
        if (visitingPredicate && node.getExpression() != null && !isAccessUsingClass(node.getExpression())) {
            createIfNullReturnPredicateId(node.getExpression());
        }
        return true;
    }

    @Override
    public boolean visit(MethodDeclaration node) {
        if (CoCoUtil.isPredicate(node)) {
            visitingPredicate = true;
            predicateId = CoCoUtil.getPredicateId(node);
            statements = new LinkedList<Statement>();
        }
        return true;
    }

    @SuppressWarnings("unchecked")
    @Override
    public void endVisit(MethodDeclaration node) {
        if (visitingPredicate) {
            Block methodBody = node.getBody();
            methodBody.statements().addAll(0, statements);
            methodBody.delete();
            Block newMethodBody = ast().newBlock();
            TryStatement tryFinally = CoCoASTBuilder.createTryFinallyIfInPredicateReturn(ast(), methodBody, MINUS_TOKEN
                    + predicateId);
            newMethodBody.statements().add(tryFinally);
            node.setBody(newMethodBody);
        }
        visitingPredicate = false;
    }

    @Override
    public boolean visit(ArrayAccess node) {
        if (visitingPredicate) {
            statements.add(0, createIfIOBReturnPredicateId(node.getArray(), node.getIndex()));
        }
        return true;
    }

    // PRIVATE

    private Statement createIfIOBReturnPredicateId(Expression arrayExpression, Expression indexExpression) {
        FieldAccess lengthAccess = ast().newFieldAccess();
        lengthAccess.setExpression((Expression) ASTNode.copySubtree(ast(), arrayExpression));
        lengthAccess.setName(ast().newSimpleName(LENGHT_TOKEN));
        InfixExpression ltZero = ast().newInfixExpression();
        ltZero.setOperator(InfixExpression.Operator.LESS);
        ltZero.setLeftOperand((Expression) ASTNode.copySubtree(ast(), indexExpression));
        ltZero.setRightOperand(ast().newNumberLiteral(ZERO_TOKEN));
        InfixExpression geLength = ast().newInfixExpression();
        geLength.setOperator(InfixExpression.Operator.GREATER_EQUALS);
        geLength.setLeftOperand((Expression) ASTNode.copySubtree(ast(), indexExpression));
        geLength.setRightOperand(lengthAccess);
        InfixExpression or = ast().newInfixExpression();
        or.setOperator(InfixExpression.Operator.CONDITIONAL_OR);
        or.setLeftOperand(ltZero);
        or.setRightOperand(geLength);
        ReturnStatement returnStatement = ast().newReturnStatement();
        returnStatement.setExpression(ast().newNumberLiteral(MINUS_TOKEN + predicateId));
        IfStatement ifStatement = ASTBuilder.createIfThenStatement(ast(), or, returnStatement);
        return ifStatement;
    }

    private void createIfNullReturnPredicateId(Expression expression) {
        InfixExpression isNull = createInfixExpressionIsNull((Expression) ASTNode.copySubtree(ast(), expression));
        String returnValue = MINUS_TOKEN + predicateId;
        for (Expression condition : conditions) {
            if (condition.subtreeMatch(new CoCoASTMatcher(), isNull)) {
                returnValue = MINUS_TOKEN + conditionToId.get(condition);
                break;
            }
        }
        ReturnStatement returnStatement = ast().newReturnStatement();
        returnStatement.setExpression(ast().newNumberLiteral(returnValue));
        IfStatement ifStatement = ASTBuilder.createIfThenStatement(ast(), isNull, returnStatement);
        statements.add(0, ifStatement);
    }

    private InfixExpression createInfixExpressionIsNull(Expression leftExpression) {
        InfixExpression isNull = ast().newInfixExpression();
        isNull.setOperator(InfixExpression.Operator.EQUALS);
        isNull.setRightOperand(ast().newNullLiteral());
        isNull.setLeftOperand(leftExpression);
        return isNull;
    }

    private boolean isAccessUsingClass(Expression expr) {
        if (expr instanceof Name) {
            IBinding binding = ((Name) expr).resolveBinding();
            if (binding == null) {
                // TODO: do something
                return false;
            }
            IJavaElement qualifierJavaElement = binding.getJavaElement();
            if (qualifierJavaElement instanceof ResolvedSourceType
                    || qualifierJavaElement instanceof ResolvedBinaryType) {
                return true;
            }
        }
        return false;
    }
}
