package coco.task.etc;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.Block;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.Initializer;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.TryStatement;

import coco.CoCoInstrumentationOpt;
import coco.CoCoASTV;
import coco.util.ASTBuilder;
import coco.util.CoCoASTBuilder;
import coco.util.CoCoUtil;

/**
 * Insert try/finally for each method body. In addition, it insert
 * touchBeginMethod/touchBeginStatic before try block and invoke
 * touchEndMethod/touchEndStatic in finally block. Note that this task must be
 * invoked last because try/finally blocks would be instrumented otherwise.
 * 
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 * @author Rohan Sharma (sharma.rohan1990@gmail.com)
 */
public class InsertMethodTouchASTV extends CoCoASTV {

    // ASTVisitor interface

    /**
     * Static blocks are seen as methods.
     */
    @Override
    public boolean visit(Initializer node) {
        Block newBody = createNewBody(CoCoInstrumentationOpt.TOUCH_BEGIN_STATIC,
                CoCoInstrumentationOpt.TOUCH_END_STATIC, node.getBody());
        node.setBody(newBody);
        return true;
    }

    /**
     * Need to keep track of beginning/end of methods/constructors.
     */
    @SuppressWarnings("unchecked")
    public boolean visit(MethodDeclaration node) {
        // ignore predicate methods and methods with no body
        if (CoCoUtil.isCondition(node) || CoCoUtil.isPredicate(node) || node.getBody() == null) {
            return false;
        }
        // skip "super" and "this" constructor invocations
        Block oldBody = node.getBody();
        if (node.isConstructor()) {
            for (Object o : oldBody.statements()) {
                if (o instanceof Block) {
                    oldBody = (Block) o;
                    break;
                }
            }
            oldBody.delete();
        }

        Block newBody = createNewBody(CoCoInstrumentationOpt.TOUCH_BEGIN_METHOD,
                CoCoInstrumentationOpt.TOUCH_END_METHOD, oldBody);
        if (node.isConstructor()) {
            node.getBody().statements().add(newBody);
        } else {
            node.setBody(newBody);
        }
        return true;
    }

    // PRIVATE

    @SuppressWarnings("unchecked")
    private Block createNewBody(String beginMethod, String endMethod, Block oldBody) {
        Block newBody = ast().newBlock();
        // add touch at the beginning of each method
        MethodInvocation touchBeginMethod = ASTBuilder.createMethodInvocation(ast(),
                CoCoASTBuilder.createInstMethodInvocation(ast()), beginMethod, new Expression[0]);
        newBody.statements().add(ast().newExpressionStatement(touchBeginMethod));
        // finally block must invoke method end touch
        MethodInvocation instInvocation = CoCoASTBuilder.createInstMethodInvocation(ast());
        MethodInvocation mi = ASTBuilder.createMethodInvocation(ast(), instInvocation, endMethod, new Expression[0]);
        Block finallyBlock = ast().newBlock();
        finallyBlock.statements().add(ast().newExpressionStatement(mi));
        // try block is the same as original body of the method
        TryStatement tryFinally = ASTBuilder.createTryFinally(ast(), (Block) ASTNode.copySubtree(ast(), oldBody),
                finallyBlock);
        // add try/finally block
        newBody.statements().add(tryFinally);
        return newBody;
    }
}
