package coco.task.etc;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IncrementalProjectBuilder;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTMatcher;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.ExpressionStatement;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jface.text.Document;
import org.eclipse.jface.text.IDocument;
import org.eclipse.text.edits.TextEdit;

import coco.CoCoASTV;
import coco.task.EmptyProgressMonitor;
import coco.task.ITask;
import coco.util.CoCoASTBuilder;
import coco.util.CoCoUtil;

/**
 * Responsible for removing predicates and touch invocations. It does NOT remove
 * inst() invocation, since it may be used as a marker in other phases. This
 * visitor is primarily used for cleaning FASE instrumentation.
 * 
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 * @author Rohan Sharma (sharma.rohan1990@gmail.com)
 */
public class RemoveExistingInstrumentationTask extends CoCoASTV implements ITask {

    /** Reference to the project that contains file(s) to be instrumented */
    private IProject project;

    /** ASTMatcher to be used */
    private ASTMatcher astMatcher = new ASTMatcher();

    /** Marker invocation to be used in matching */
    private MethodInvocation markerInvocation;

    public RemoveExistingInstrumentationTask(IProject project) {
        this.project = project;
    }

    @Override
    public void perform(ICompilationUnit icu) throws Exception {
        // TODO: remove duplicate code from here

        ASTParser parser = ASTParser.newParser(AST.JLS3);
        parser.setSource(icu);
        CompilationUnit cu = (CompilationUnit) parser.createAST(EmptyProgressMonitor.inst());
        IDocument doc = new Document(icu.getSource());
        cu.recordModifications();
        cu.accept(this);
        TextEdit textEdits = cu.rewrite(doc, null);
        textEdits.apply(doc);
        icu.getBuffer().setContents(doc.get());
        icu.save(EmptyProgressMonitor.inst(), true);
        project.build(IncrementalProjectBuilder.FULL_BUILD, EmptyProgressMonitor.inst());
    }

    // ASTVisitor interface

    @Override
    public boolean visit(MethodDeclaration node) {
        // delete a method if the method is predicate
        if (CoCoUtil.isPredicate(node)) {
            node.delete();
            return false;
        }
        return true;
    }

    @Override
    public boolean visit(ExpressionStatement node) {
        if (node.getExpression() instanceof MethodInvocation) {
            MethodInvocation mi = (MethodInvocation) node.getExpression();
            Expression inst = mi.getExpression();
            if (inst != null && inst.subtreeMatch(astMatcher, getMarkerInvocation())) {
                inst.delete();
                node.setExpression(inst);
            }
        }
        return true;
    }

    // PRIVATE

    // optimization
    private MethodInvocation getMarkerInvocation() {
        if (markerInvocation == null) {
            markerInvocation = CoCoASTBuilder.createInstMethodInvocation(ast());
        }
        return markerInvocation;
    }

}
