package coco.task.extract.pct;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.AnonymousClassDeclaration;
import org.eclipse.jdt.core.dom.Block;
import org.eclipse.jdt.core.dom.ClassInstanceCreation;
import org.eclipse.jdt.core.dom.ConditionalExpression;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.InfixExpression;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.Modifier.ModifierKeyword;
import org.eclipse.jdt.core.dom.PrimitiveType;
import org.eclipse.jdt.core.dom.ReturnStatement;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.Statement;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;

import coco.CoCoASTV;
import coco.IResettableV;
import coco.util.CoCoASTMatcher;
import coco.util.CoCoUtil;

/**
 * Analyzes code, finds extracted <b>conditions</b> and makes predicates from
 * conditions. It does not add code into the unit (@see
 * {@link AddPredicatesASTV}).
 * 
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 * @author Rohan Sharma (sharma.rohan1990@gmail.com)
 */
public class ConstructPredicatesASTV extends CoCoASTV implements IResettableV {

    /** Next id to be used for predicate */
    private int nextPredicateId;

    /** Mapping type to predicates in scope */
    private Map<Integer, List<MethodDeclaration>> typeToPredicates = new HashMap<Integer, List<MethodDeclaration>>();

    /** Mapping type to expression in return statement of conditions */
    private Map<Integer, List<Expression>> typeSeqToConditions = new HashMap<Integer, List<Expression>>();

    /** mapping condition to conditionId */
    private Map<Expression, String> conditionToId = new HashMap<Expression, String>();

    /** current type hierarchy */
    private Stack<Integer> typeIds = new Stack<Integer>();

    /** next type in order */
    private int nextTypeInSeq = 0;

    public ConstructPredicatesASTV(int nextPredicateId) {
        this.nextPredicateId = nextPredicateId;
    }

    public int getNextPredicateId() {
        return nextPredicateId;
    }

    @Override
    public void reset() {
        nextTypeInSeq = 0;
    }

    private final void addNewType() {
        nextTypeInSeq++;
        if (!typeSeqToConditions.containsKey(nextTypeInSeq)) {
            typeSeqToConditions.put(nextTypeInSeq, new LinkedList<Expression>());
            typeToPredicates.put(nextTypeInSeq, new LinkedList<MethodDeclaration>());
        }
        typeIds.push(nextTypeInSeq);
    }

    // ASTVisitor interface

    @Override
    public final boolean visit(AnonymousClassDeclaration node) {
        addNewType();
        return true;
    }

    @Override
    public final void endVisit(AnonymousClassDeclaration node) {
        typeIds.pop();
    }

    @Override
    public boolean oVisit(TypeDeclaration node) {
        addNewType();
        return true;
    }

    @Override
    public void oEndVisit(TypeDeclaration node) {
        typeIds.pop();
    }

    @Override
    public boolean visit(MethodDeclaration node) {
        if (CoCoUtil.isPredicate(node)) {
            return false;
        }

        if (CoCoUtil.isCondition(node)) {
            Expression nodeExpression = ((ReturnStatement) (node.getBody().statements().get(node.getBody().statements()
                    .size() - 1))).getExpression();
            if (nodeExpression instanceof ClassInstanceCreation) {
                return true;
            }

            if (!node.getReturnType2().isPrimitiveType()) {
                InfixExpression eqNull = ast().newInfixExpression();
                eqNull.setOperator(InfixExpression.Operator.EQUALS);
                eqNull.setRightOperand(ast().newNullLiteral());
                eqNull.setLeftOperand((Expression) ASTNode.copySubtree(ast(), nodeExpression));
                nodeExpression = eqNull;
            }

            List<Expression> conditionalExpressions = typeSeqToConditions.get(typeIds.peek());
            Expression earlierCondition = null;
            for (Expression condition : conditionalExpressions) {
                if (condition.subtreeMatch(new CoCoASTMatcher(), nodeExpression)) {
                    earlierCondition = condition;
                    break;
                }
            }
            List<MethodDeclaration> predicates = typeToPredicates.get(typeIds.peek());
            if (earlierCondition == null) {
                conditionalExpressions.add(nodeExpression);
                conditionToId.put(nodeExpression, Integer.toString(nextPredicateId));
                predicates.add(createPredicateDeclaration(ast(), node, Integer.toString(nextPredicateId)));
                nextPredicateId++;
            } else {
                // numOfIsomorphicPredicates++;
                // predicates.add(createPredicateDeclarationForward(ast, node,
                // earlierCondition));
            }
            // TODO: here we add original predicate since we may depend on it
            // (it is
            // not nice approach)
            // predicates.add(node);
        }
        return true;
    }

    public Map<Integer, List<MethodDeclaration>> getTypeToPredicates() {
        return typeToPredicates;
    }

    public Map<Integer, List<Expression>> getTypeSeqToConditions() {
        return typeSeqToConditions;
    }

    public Map<Expression, String> getConditionToId() {
        return conditionToId;
    }

    @SuppressWarnings("unchecked")
    private MethodDeclaration createPredicateDeclaration(AST ast, MethodDeclaration condition, String predicateId) {
        MethodDeclaration predicate = ast.newMethodDeclaration();
        for (Object obj : condition.parameters()) {
            SingleVariableDeclaration svd = (SingleVariableDeclaration) obj;
            predicate.parameters().add(ASTNode.copySubtree(ast, svd));
        }
        predicate.setName(ast.newSimpleName(CoCoUtil.makePredicateString(predicateId)));
        predicate.modifiers().add(ast.newModifier(ModifierKeyword.PUBLIC_KEYWORD));
        if (Modifier.isStatic(condition.getModifiers())) {
            predicate.modifiers().add(ast.newModifier(ModifierKeyword.STATIC_KEYWORD));
        }
        for (Object obj : condition.typeParameters()) {
            predicate.typeParameters().add(ASTNode.copySubtree(ast, (ASTNode) obj));
        }
        predicate.setReturnType2(ast.newPrimitiveType(PrimitiveType.INT));
        List<Statement> statements = new LinkedList<Statement>();
        List<Statement> originalStatements = (List<Statement>) condition.getBody().statements();
        for (int i = 0; i < originalStatements.size(); i++) {
            if (i == originalStatements.size() - 1) {
                ConditionalExpression conditionalExpression = ast.newConditionalExpression();
                conditionalExpression.setThenExpression(ast.newNumberLiteral(predicateId));
                conditionalExpression.setElseExpression(ast.newNumberLiteral("-" + predicateId));
                Expression originalExpression = ((Expression) ASTNode.copySubtree(ast,
                        ((ReturnStatement) originalStatements.get(i)).getExpression()));
                if (!condition.getReturnType2().isPrimitiveType()) {
                    // NPE. check with null
                    InfixExpression eqNull = ast.newInfixExpression();
                    eqNull.setOperator(InfixExpression.Operator.EQUALS);
                    eqNull.setLeftOperand(originalExpression);
                    eqNull.setRightOperand(ast.newNullLiteral());
                    originalExpression = eqNull;
                }
                VariableDeclarationFragment tmpResult = ast.newVariableDeclarationFragment();
                tmpResult.setName(ast().newSimpleName("tmpResult"));
                VariableDeclarationStatement tmpResultStmt = ast.newVariableDeclarationStatement(tmpResult);
                tmpResultStmt.setType(ast.newPrimitiveType(PrimitiveType.BOOLEAN));
                tmpResult.setInitializer(originalExpression);
                statements.add(tmpResultStmt);
                // ParenthesizedExpression parenthesis =
                // ast().newParenthesizedExpression();
                // parenthesis.setExpression(originalExpression);
                conditionalExpression.setExpression(ast.newSimpleName("tmpResult"));
                ReturnStatement returnStatement = ast.newReturnStatement();
                returnStatement.setExpression(conditionalExpression);
                statements.add(returnStatement);
            } else {
                statements.add((Statement) ASTNode.copySubtree(ast, originalStatements.get(i)));
            }
        }
        Block predicateBody = ast.newBlock();
        predicateBody.statements().addAll(statements);
        predicate.setBody(predicateBody);
        return predicate;
    }
}
