package coco.task.extract.pct;

import org.eclipse.jdt.core.dom.AssertStatement;
import org.eclipse.jdt.core.dom.ConditionalExpression;
import org.eclipse.jdt.core.dom.DoStatement;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.ForStatement;
import org.eclipse.jdt.core.dom.IfStatement;
import org.eclipse.jdt.core.dom.InfixExpression;
import org.eclipse.jdt.core.dom.ParenthesizedExpression;
import org.eclipse.jdt.core.dom.WhileStatement;

/**
 * Extract conditions from conditional statements. Conditions are split to
 * atomic conditions (e.g., for "left && right", two conditions are created).
 * 
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 * @author Rohan Sharma (sharma.rohan1990@gmail.com)
 * 
 * @see DetectCompositeConditionsASTV
 */
class DetectAtomicConditionsASTV extends DetectConditionsASTV {

    public DetectAtomicConditionsASTV(DetectPureMethodsASTV pureVisitor) {
        super(pureVisitor);
    }

    // ASTVisitor interface

    @Override
    public void endVisit(InfixExpression node) {
        if (isCompositeExpression(node)) {
            visitAtomicPredicate(node.getLeftOperand());
            visitAtomicPredicate(node.getRightOperand());
            if (node.hasExtendedOperands())
                for (Object operand : node.extendedOperands())
                    visitAtomicPredicate((Expression) operand);
        }
    }

    @Override
    public void endVisit(ParenthesizedExpression node) {
        Expression expr = node.getExpression();
        if (expr instanceof InfixExpression && isSimpleBooleanExpression((InfixExpression) expr))
            visitCondition(expr);
    }

    @Override
    public void endVisit(AssertStatement node) {
        visitAtomicPredicate(node.getExpression());
    }

    @Override
    public void endVisit(DoStatement node) {
        visitAtomicPredicate(node.getExpression());
    }

    @Override
    public void endVisit(ForStatement node) {
        if (node.getExpression() != null) {
            visitAtomicPredicate(node.getExpression());
        }
    }

    @Override
    public void endVisit(IfStatement node) {
        visitAtomicPredicate(node.getExpression());
    }

    @Override
    public void endVisit(WhileStatement node) {
        visitAtomicPredicate(node.getExpression());
    }

    @Override
    public void endVisit(ConditionalExpression node) {
        visitAtomicPredicate(node.getExpression());
    }

    // PRIVATE

    private void visitAtomicPredicate(Expression expr) {
        if (!(expr instanceof ParenthesizedExpression)
                && (!(expr instanceof InfixExpression) || !isCompositeExpression((InfixExpression) expr)))
            visitCondition(expr);
    }

    private boolean isCompositeExpression(InfixExpression expr) {
        return expr.getOperator() == InfixExpression.Operator.CONDITIONAL_AND
                || expr.getOperator() == InfixExpression.Operator.CONDITIONAL_OR;
    }

    private boolean isSimpleBooleanExpression(InfixExpression expr) {
        return expr.getOperator() == InfixExpression.Operator.EQUALS
                || expr.getOperator() == InfixExpression.Operator.NOT_EQUALS
                || expr.getOperator() == InfixExpression.Operator.LESS
                || expr.getOperator() == InfixExpression.Operator.LESS_EQUALS
                || expr.getOperator() == InfixExpression.Operator.GREATER
                || expr.getOperator() == InfixExpression.Operator.GREATER_EQUALS;
    }

}
