package coco.task.extract.pct;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.jdt.core.Flags;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.ASTVisitor;
import org.eclipse.jdt.core.dom.AnonymousClassDeclaration;
import org.eclipse.jdt.core.dom.Assignment;
import org.eclipse.jdt.core.dom.ClassInstanceCreation;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.PostfixExpression;
import org.eclipse.jdt.core.dom.PrefixExpression;
import org.eclipse.jdt.core.dom.TypeDeclaration;

import coco.util.CoCoUtil;

/**
 * Finds pure methods. Pure method is simply defined as: non-abstract method
 * without: assignment(s), ++/--, method calls to non-pure methods (method calls
 * to other types are considered non-pure).
 * 
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 * @author Rohan Sharma (sharma.rohan1990@gmail.com)
 */
public class DetectPureMethodsASTV extends ASTVisitor {

    private final Map<String, List<MethodDeclaration>> typeToPureMethods;
    private final Map<String, List<String>> typeToSignatures;
    private String currentType;

    private final SingleMethodASTVisitor methodASTVisitor;
    private final CallGraphASTVisitor callGraphASTVisitor;

    /**
     * It should be invoked on each method of a type.
     */
    class SingleMethodASTVisitor extends ASTVisitor {
        boolean isPure;

        public boolean isPure() {
            return isPure;
        }

        // method to be used when condition is detected to reset state
        public void setCurrentType(String t) {
            currentType = t;
            isPure = true;
        }

        @Override
        public boolean visit(AnonymousClassDeclaration node) {
            // ignore anonymous classes
            return false;
        }

        @Override
        public boolean visit(MethodDeclaration node) {
            // if method is abstract => not pure
            isPure = !(Flags.isAbstract(node.getFlags()) || Modifier.isAbstract(node.getModifiers()));
            return isPure;
        }

        @Override
        public boolean visit(ClassInstanceCreation node) {
            // if object is created => not pure
            isPure = false;
            return true;
        }

        @Override
        public boolean visit(MethodInvocation node) {
            IType type = (IType) node.resolveMethodBinding().getDeclaringClass().getJavaElement();
            try {
                // if method is invoked on another type => not pure
                String key = type.getFullyQualifiedName() + type.getSourceRange().getOffset();
                if (!currentType.equals(key)) {
                    isPure = false;
                }
            } catch (JavaModelException e) {
                e.printStackTrace();
            }
            return true;
        }

        @Override
        public boolean visit(Assignment node) {
            // if there is an assignment => not pure
            isPure = false;
            return true;
        }

        @Override
        public boolean visit(PrefixExpression node) {
            // if ++v or --v => not pure
            isPure = !(node.getOperator() == PrefixExpression.Operator.INCREMENT || node.getOperator() == PrefixExpression.Operator.DECREMENT);
            return true;
        }

        @Override
        public boolean visit(PostfixExpression node) {
            // if v++ or v-- => not pure
            isPure = !(node.getOperator() == PostfixExpression.Operator.INCREMENT || node.getOperator() == PostfixExpression.Operator.DECREMENT);
            return true;
        }
    }

    public class CallGraphASTVisitor extends SingleMethodASTVisitor {
        @Override
        public boolean visit(MethodInvocation node) {
            super.visit(node);
            if (!isPure) {
                return true;
            }
            IType type = (IType) node.resolveMethodBinding().getDeclaringClass().getJavaElement();
            try {
                String key = type.getFullyQualifiedName() + type.getSourceRange().getOffset();
                if (currentType.equals(key)) {
                    String signature = CoCoUtil.getSignature(node.resolveMethodBinding());
                    if (!typeToSignatures.get(currentType).contains(signature)) {
                        isPure = false;
                    }
                }
            } catch (JavaModelException e) {
                e.printStackTrace();
            }
            return true;
        }
    }

    public DetectPureMethodsASTV() {
        this.typeToPureMethods = new HashMap<String, List<MethodDeclaration>>();
        this.typeToSignatures = new HashMap<String, List<String>>();
        this.methodASTVisitor = new SingleMethodASTVisitor();
        this.callGraphASTVisitor = new CallGraphASTVisitor();
    }

    public CallGraphASTVisitor getCallGraphASTVisitor() {
        return callGraphASTVisitor;
    }

    @Override
    public boolean visit(AnonymousClassDeclaration node) {
        IType type = (IType) node.resolveBinding().getJavaElement();
        try {
            currentType = type.getFullyQualifiedName() + type.getSourceRange().getOffset();
            typeToPureMethods.put(currentType, new ArrayList<MethodDeclaration>());
            typeToSignatures.put(currentType, new ArrayList<String>());
            checkSingleMethods(CoCoUtil.getMethods(node));
            checkCallGraph(typeToPureMethods.get(currentType), typeToSignatures.get(currentType));
        } catch (JavaModelException e) {
            e.printStackTrace();
        }
        return true;
    }

    @Override
    public boolean visit(TypeDeclaration node) {
        IType type = (IType) node.resolveBinding().getJavaElement();
        try {
            currentType = type.getFullyQualifiedName() + type.getSourceRange().getOffset();
            typeToPureMethods.put(currentType, new ArrayList<MethodDeclaration>());
            typeToSignatures.put(currentType, new ArrayList<String>());
            checkSingleMethods(node.getMethods());
            checkCallGraph(typeToPureMethods.get(currentType), typeToSignatures.get(currentType));

            // System.out.println(typeToSignatures.get(currentType));
        } catch (JavaModelException e) {
            e.printStackTrace();
        }
        return true;
    }

    private void checkSingleMethods(MethodDeclaration[] methods) {
        for (MethodDeclaration md : methods) {
            md.accept(methodASTVisitor);
            if (md.resolveBinding() != null && methodASTVisitor.isPure) {
                typeToPureMethods.get(currentType).add(md);
                typeToSignatures.get(currentType).add(CoCoUtil.getSignature(md));
            }
        }
    }

    private void checkCallGraph(List<MethodDeclaration> methods, List<String> signatures) {
        int size;
        do {
            size = methods.size();
            Iterator<MethodDeclaration> iter = methods.iterator();
            Iterator<String> iter2 = signatures.iterator();
            while (iter.hasNext()) {
                MethodDeclaration md = iter.next();
                iter2.next();
                md.accept(callGraphASTVisitor);
                if (!callGraphASTVisitor.isPure) {
                    iter.remove();
                    iter2.remove();
                }
            }
        } while (size != methods.size());
    }
}
