package coco.task.extract.pct;

import java.lang.reflect.Modifier;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.internal.corext.refactoring.code.ExtractMethodRefactoring;
import org.eclipse.ltk.core.refactoring.Change;
import org.eclipse.ltk.core.refactoring.Refactoring;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;

import coco.CoCoInstrumentationOpt;
import coco.log.Log;
import coco.task.EmptyProgressMonitor;
import coco.task.ITask;
import coco.task.PerformRefactoring;
import coco.task.etc.AddNullIOBChecksASTV;
import coco.util.ASTVHelper;
import coco.util.CoCoUtil;
import coco.util.ICUCompiler;
import coco.util.ICUCompiler.ICUCompilerResult;

/**
 * Task to extract predicates from assertions, conditional statements and
 * runtime checks. This task also adds extracted predicates in ICUs. This task
 * is implemented to perform the following steps.
 * 
 * <ul>
 * <li>Find pure methods</li>
 * <li>Find/construct predicates from runtime checks</li>
 * <li>Find/construct predicates from conditional statements</li>
 * <li>Add predicates in ICU</li>
 * <li>Add checks for null dereference and IOB in predicate methods</li>
 * </ul>
 * 
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 * @author Rohan Sharma (sharma.rohan1990@gmail.com)
 */
@SuppressWarnings("restriction")
public class ExtractPredicatesTask implements ITask {

    /** Command line arguments */
    private final CoCoInstrumentationOpt args;

    /** Project that contains icu */
    private final IProject project;

    /** Parser reused for traversal/visit */
    private final ASTParser astParser;

    /**
     * Number of predicates that is extracted (note that this number may differ
     * from number of predicates)
     */
    private int numOfExtractedPredicates;

    private boolean isFirstTime;

    public ExtractPredicatesTask(CoCoInstrumentationOpt args, IProject project) {
        this.args = args;
        this.project = project;
        this.astParser = ASTParser.newParser(AST.JLS3);
        this.isFirstTime = true;
    }

    public int getNumOfExtractedPredicates() {
        return numOfExtractedPredicates;
    }

    @Override
    public void perform(ICompilationUnit icu) throws Exception {
        // find pure methods
        DetectPureMethodsASTV detectPureMethodsV = new DetectPureMethodsASTV();
        ASTVHelper.acceptVisitor(project, icu, detectPureMethodsV, true);

        ConstructPredicatesASTV constructPredicatesV;
        if (args.isContinuousPredicateCounting()) {
            if (isFirstTime) {
                isFirstTime = false;
                numOfExtractedPredicates = args.getBasePredicateId() - 1;
            }
            constructPredicatesV = new ConstructPredicatesASTV(numOfExtractedPredicates + 1);
        } else {
            constructPredicatesV = new ConstructPredicatesASTV(args.getBasePredicateId());
        }
        int nextPredicateIdBeforeVisit = constructPredicatesV.getNextPredicateId();

        // extract predicates from field/method accesses
        DetectNullChecksASTV extractNullChecks = new DetectNullChecksASTV(detectPureMethodsV);
        extractCodeToFormPredicates(extractNullChecks, icu, constructPredicatesV);

        // extract predicates from conditional statements
        DetectConditionsASTV extractConditionsV;
        if (args.isAtomicPredicates.g()) {
            extractConditionsV = new DetectAtomicConditionsASTV(detectPureMethodsV);
        } else {
            extractConditionsV = new DetectCompositeConditionsASTV(detectPureMethodsV);
        }
        extractCodeToFormPredicates(extractConditionsV, icu, constructPredicatesV);

        // add predicates in ICU
        AddPredicatesASTV addPredicatesV = new AddPredicatesASTV(constructPredicatesV);
        ASTVHelper.acceptVisitor(project, icu, addPredicatesV, false);

        // add checks in predicates for null and IOB
        AddNullIOBChecksASTV addRuntimeChecksV = new AddNullIOBChecksASTV(constructPredicatesV);
        ASTVHelper.acceptVisitor(project, icu, addRuntimeChecksV, true);

        numOfExtractedPredicates += constructPredicatesV.getNextPredicateId() - nextPredicateIdBeforeVisit;
    }

    private void extractCodeToFormPredicates(DetectXASTV extractX, ICompilationUnit icu,
            ConstructPredicatesASTV constructPredicatesV) throws Exception {
        while (true) {
            astParser.setResolveBindings(true);
            astParser.setSource(icu);
            ASTNode root = astParser.createAST(EmptyProgressMonitor.inst());
            root.accept(extractX);
            if (!extractX.hasCondition()) {
                break;
            }
            Refactoring refactoring = getExtractMethodRefactoring(icu, extractX);
            if (refactoring == null) {
                extractX.reset();
                continue;
            }
            PerformRefactoring.performRefactoring(refactoring, true);
            Change undoChange = PerformRefactoring.getUndoChange();
            project.refreshLocal(IResource.DEPTH_INFINITE, EmptyProgressMonitor.inst());
            extractX.reset();

            ICUCompilerResult compilerResult = ICUCompiler.inst().compile(icu);
            if (!compilerResult.hasErrorOtherThan("Unreachable code", "This method must return a result of type")) {
                // if (!compilerResult
                // .hasMatchingError("Cannot make a static reference to the non-static type")
                // && !compilerResult
                // .hasMatchingError("static methods can only be declared in a static or top level type")
                // && !compilerResult
                // .hasMatchingError("is not a valid substitute for the bounded parameter"))
                // {
                astParser.setSource(icu);
                CompilationUnit cu = (CompilationUnit) astParser.createAST(EmptyProgressMonitor.inst());
                constructPredicatesV.reset();
                cu.accept(constructPredicatesV);
            } else {
                Log.e("ExtractPredicatesTask:: ERROR:: " + compilerResult.getErrors().toString());
            }

            PerformRefactoring.undoRefactoring(undoChange);
        }
    }

    private Refactoring getExtractMethodRefactoring(ICompilationUnit icu, DetectXASTV extractXV) throws Exception {
        astParser.setResolveBindings(true);
        astParser.setSource(icu);
        CompilationUnit cu = (CompilationUnit) astParser.createAST(EmptyProgressMonitor.inst());
        ExtractMethodRefactoring refactoring = new ExtractMethodRefactoring(cu, extractXV.getStartPosition(),
                extractXV.getLength());
        refactoring.setMethodName(CoCoUtil.makeConditionString(extractXV.getCurrentConditionId()));
        refactoring.setVisibility(Modifier.PUBLIC);
        RefactoringStatus status = refactoring.checkInitialConditions(EmptyProgressMonitor.inst());
        if (status.hasFatalError()) {
            Log.e(getClass() + ":: refactoring");
            return null;
        }
        refactoring.setDestination(0);
        refactoring.setReplaceDuplicates(false);
        return refactoring;
    }

}
