package coco.task.instantiate;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;

import coco.log.Log;

/**
 * Keeps mapping from each type to predicates that the type declares. TODO: Do
 * we have to use our own structure?
 * 
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 * @author Rohan Sharma (sharma.rohan1990@gmail.com)
 */
public class ITypePredicatesMap {

    private static class Node {
        @SuppressWarnings("unused")
        final int level;
        final IType type;
        final List<IMethod> predicates;

        Node(int level, IType type) {
            this.level = level;
            this.type = type;
            predicates = new ArrayList<IMethod>();
        }
    }

    private int currLevel;

    private final Map<String, Node> nodes;

    public ITypePredicatesMap() {
        this.nodes = new HashMap<String, ITypePredicatesMap.Node>();
    }

    /**
     * Starts a new level. Each new type should start a level.
     */
    public void beginLevel() {
        currLevel++;
    }

    /**
     * Ends a level. End of a type should end a level.
     */
    public void endLevel() {
        currLevel--;
    }

    /**
     * Creates a mapping from the given type to the given predicate.
     */
    public void add(IType t, IMethod m) {
        String key = t.getFullyQualifiedName();
        Log.m(getClass() + ":: ADDED PREDICATE:: " + key + " " + m.getElementName());
        try {
            key += t.getSourceRange().getOffset();
        } catch (JavaModelException e) {
            throw new RuntimeException("Cannot obtain source range (full name) for a type.");
        }
        if (!nodes.containsKey(key)) {
            nodes.put(key, new Node(currLevel, t));
        }
        nodes.get(key).predicates.add(m);
    }

    /**
     * Finds all the predicates that are available for the given type. The
     * predicates include all predicates in the type plus all the predicates
     * from higher level/types.
     * 
     * TODO: we may want to (1) include only parents of the given type and the
     * type itself or (2) all the types that are on higher lexical scope.
     */
    public Map<IMethod, IType> getPredicates(IType t) {
        String key = t.getFullyQualifiedName();
        try {
            key += t.getSourceRange().getOffset();
        } catch (JavaModelException e) {
            throw new RuntimeException("Cannot obtain source range (full name) for a type.");
        }
        Map<IMethod, IType> allPredicates = new HashMap<IMethod, IType>();
        Node selectedNode = nodes.get(key);
        if (selectedNode != null) {
            // the following code includes all predicates from types defined on
            // higher lexical scope
            // for (Entry<String, Node> n : nodes.entrySet()) {
            // if (n.getValue().level < selectedNode.level) {
            // for (IMethod m : n.getValue().predicates) {
            // allPredicates.put(m, n.getValue().type);
            // }
            // }
            // }
            for (IMethod m : selectedNode.predicates) {
                allPredicates.put(m, selectedNode.type);
            }
        }
        return allPredicates;
    }

    /**
     * Returns number of predicates in the structure.
     */
    public int size() {
        int size = 0;
        for (Entry<String, Node> n : nodes.entrySet()) {
            size += n.getValue().predicates.size();
        }
        return size;
    }
}
