package coco.task.instantiate;

import java.util.Iterator;
import java.util.LinkedList;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.Block;
import org.eclipse.jdt.core.dom.CatchClause;
import org.eclipse.jdt.core.dom.ConstructorInvocation;
import org.eclipse.jdt.core.dom.DoStatement;
import org.eclipse.jdt.core.dom.EnhancedForStatement;
import org.eclipse.jdt.core.dom.ForStatement;
import org.eclipse.jdt.core.dom.IfStatement;
import org.eclipse.jdt.core.dom.LabeledStatement;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.Statement;
import org.eclipse.jdt.core.dom.SuperConstructorInvocation;
import org.eclipse.jdt.core.dom.SwitchCase;
import org.eclipse.jdt.core.dom.SwitchStatement;
import org.eclipse.jdt.core.dom.TryStatement;
import org.eclipse.jdt.core.dom.WhileStatement;

import coco.CoCoASTV;
import coco.util.CoCoUtil;

/**
 * Inserts new {@link Block} statement at the beginning of each basic block.
 * Important: endVisit cannot be used because blocks/parents are needed for
 * inner nodes.
 * 
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 * @author Rohan Sharma (sharma.rohan1990@gmail.com)
 */
@SuppressWarnings("unchecked")
public class InsertBasicBlockASTV extends CoCoASTV {

    private int currentLevel;
    private int globalLevel;
    private boolean hasNextLevel;

    public void advance() {
        currentLevel = 0;
        globalLevel++;
        hasNextLevel = false;
    }

    public boolean hasNextLevel() {
        return hasNextLevel;
    }

    private void insertAfter(ASTNode node) {
        Block parent;
        // if the parent is labeled statement insert after it
        if (node.getParent() instanceof LabeledStatement) {
            parent = (Block) node.getParent().getParent();
            // node has to be faked! (see use of node below)
            node = node.getParent();
        } else {
            parent = (Block) node.getParent();
        }
        int index = parent.statements().indexOf(node);
        parent.statements().add(index + 1, ast().newBlock());
    }

    // ASTVisitor interface

    @Override
    public boolean visit(Block node) {
        currentLevel++;
        if (currentLevel == globalLevel) {
            hasNextLevel = true;
        }
        return true;
    }

    @Override
    public void endVisit(Block node) {
        currentLevel--;
    }

    @Override
    public boolean visit(MethodDeclaration node) {
        if (CoCoUtil.isCondition(node) || CoCoUtil.isPredicate(node) || node.getBody() == null) {
            return false;
        }
        currentLevel++;
        if (currentLevel == globalLevel) {
            hasNextLevel = true;
        } else {
            return true;
        }
        Block inner = ast().newBlock();
        Iterator<Object> iter = node.getBody().statements().iterator();
        while (iter.hasNext()) {
            Statement s = (Statement) iter.next();
            if (!(s instanceof SuperConstructorInvocation || s instanceof ConstructorInvocation)) {
                iter.remove();
                s.delete();
                inner.statements().add(s);
            }
        }
        node.getBody().statements().add(inner);
        return true;
    }

    @Override
    public boolean visit(IfStatement node) {
        if (currentLevel != globalLevel)
            return true;
        Statement thenStatement = node.getThenStatement();
        Block newThenStatement = ast().newBlock();
        node.setThenStatement(newThenStatement);
        thenStatement.delete();
        newThenStatement.statements().add(thenStatement);
        if (node.getElseStatement() == null) {
            node.setElseStatement(ast().newEmptyStatement());
        }
        Statement elseStatement = node.getElseStatement();
        Block newElseStatement = ast().newBlock();
        node.setElseStatement(newElseStatement);
        elseStatement.delete();
        newElseStatement.statements().add(elseStatement);
        insertAfter(node);
        return true;
    }

    @Override
    public boolean visit(ForStatement node) {
        if (currentLevel != globalLevel)
            return true;
        Statement body = node.getBody();
        Block newBody = ast().newBlock();
        node.setBody(newBody);
        body.delete();
        newBody.statements().add(body);
        insertAfter(node);
        return true;
    }

    @Override
    public boolean visit(EnhancedForStatement node) {
        if (currentLevel != globalLevel)
            return true;
        Statement body = node.getBody();
        Block newBody = ast().newBlock();
        node.setBody(newBody);
        body.delete();
        newBody.statements().add(body);
        insertAfter(node);
        return true;
    }

    @Override
    public boolean visit(WhileStatement node) {
        if (currentLevel != globalLevel)
            return true;
        Statement body = node.getBody();
        Block newBody = ast().newBlock();
        node.setBody(newBody);
        body.delete();
        newBody.statements().add(body);
        insertAfter(node);
        return true;
    }

    @Override
    public boolean visit(DoStatement node) {
        if (currentLevel != globalLevel)
            return true;
        Statement body = node.getBody();
        Block newBody = ast().newBlock();
        node.setBody(newBody);
        body.delete();
        newBody.statements().add(body);
        insertAfter(node);
        return true;
    }

    @Override
    public boolean visit(CatchClause node) {
        if (currentLevel != globalLevel)
            return true;
        Statement body = node.getBody();
        Block newBody = ast().newBlock();
        node.setBody(newBody);
        body.delete();
        newBody.statements().add(body);
        return true;
    }

    @Override
    public boolean visit(TryStatement node) {
        if (currentLevel != globalLevel)
            return true;
        Statement body = node.getBody();
        Block newBody = ast().newBlock();
        node.setBody(newBody);
        body.delete();
        newBody.statements().add(body);
        if (node.getFinally() == null) {
            node.setFinally(ast().newBlock());
        }
        Statement finallyBlock = node.getFinally();
        Block newFinallyBlock = ast().newBlock();
        node.setFinally(newFinallyBlock);
        finallyBlock.delete();
        newFinallyBlock.statements().add(finallyBlock);
        insertAfter(node);
        return true;
    }

    // We do not add blocks for LabeledStatement because "break/continue LABEL"
    // would not work. The solution is to enclose LabelStatement in a block that
    // can be used for instantiating predicates (see visitor that instantiates
    // predicates).
    // @Override
    // public boolean visit(LabeledStatement node) {
    // if (currentLevel != globalLevel) return true;
    // Statement body = node.getBody();
    // Block newBody = ast().newBlock();
    // node.setBody(newBody);
    // body.delete();
    // newBody.statements().add(body);
    // return true;
    // }

    @Override
    public boolean visit(SwitchStatement node) {
        if (currentLevel != globalLevel)
            return true;
        LinkedList<Statement> statements = new LinkedList<Statement>();
        for (int i = 0; i < node.statements().size(); i++) {
            Statement s = (Statement) node.statements().get(i);
            if (s instanceof SwitchCase) {
                statements.add((Statement) ASTNode.copySubtree(ast(), s));
                statements.add(ast().newBlock());
            } else {
                ((Block) statements.getLast()).statements().add((Statement) ASTNode.copySubtree(ast(), s));
            }
        }
        node.statements().clear();
        node.statements().addAll(statements);
        insertAfter(node);
        return true;
    }

}
