package coco.task.instantiate;

import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Stack;

import org.eclipse.jdt.core.Flags;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.AnonymousClassDeclaration;
import org.eclipse.jdt.core.dom.Assignment;
import org.eclipse.jdt.core.dom.Block;
import org.eclipse.jdt.core.dom.CatchClause;
import org.eclipse.jdt.core.dom.EnhancedForStatement;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.ForStatement;
import org.eclipse.jdt.core.dom.Initializer;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.Modifier.ModifierKeyword;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.Statement;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.internal.core.LocalVariable;

import coco.CoCoInstrumentationOpt;
import coco.CoCoASTV;
import coco.util.ASTBuilder;
import coco.util.CoCoASTBuilder;
import coco.util.CoCoUtil;

/**
 * Top of hierarchy that instantiate predicates at appropriate points.
 * 
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 * @author Rohan Sharma (sharma.rohan1990@gmail.com)
 */
@SuppressWarnings("restriction")
public abstract class InstantiateASTV extends CoCoASTV {

    private static final String SCOPE_DELIMITER = "*";

    private final Stack<MethodDeclaration> currentMethodDeclaration;
    private final Stack<String> variablesInScope;
    private final Stack<IType> typeScope;

    private final ITypePredicatesMap typePredicateMap;

    protected long pointId;

    public InstantiateASTV(CoCoInstrumentationOpt args, ITypePredicatesMap typePredicateMap, long pointId) {
        this.typePredicateMap = typePredicateMap;
        this.currentMethodDeclaration = new Stack<MethodDeclaration>();
        this.variablesInScope = new Stack<String>();
        this.typeScope = new Stack<IType>();
        this.pointId = pointId;
    }

    @SuppressWarnings("unchecked")
    @Override
    public final boolean visit(Initializer node) {
        // adapt static block to method declaration (see use of method
        // declarations)
        MethodDeclaration methodDeclAdapter = ast().newMethodDeclaration();
        methodDeclAdapter.modifiers().add(ast().newModifier(ModifierKeyword.STATIC_KEYWORD));
        currentMethodDeclaration.push(methodDeclAdapter);
        scope();
        return true;
    }

    @Override
    public final void endVisit(Initializer node) {
        currentMethodDeclaration.pop();
        endScope();
    }

    @Override
    public final boolean visit(AnonymousClassDeclaration node) {
        scope();
        IType type = (IType) node.resolveBinding().getJavaElement();
        typeScope.push(type);
        return true;
    }

    @Override
    public final void endVisit(AnonymousClassDeclaration node) {
        endScope();
        typeScope.pop();
    }

    @Override
    public final boolean oVisit(TypeDeclaration node) {
        scope();
        IType type = (IType) node.resolveBinding().getJavaElement();
        typeScope.push(type);
        return true;
    }

    @Override
    public final void oEndVisit(TypeDeclaration node) {
        endScope();
        typeScope.pop();
    }

    @Override
    public final boolean visit(ForStatement node) {
        scope();
        for (Object obj : node.initializers()) {
            Expression initializer = (Expression) obj;
            initializer.accept(this);
        }
        extendVisit(node);
        return true;
    }

    public void extendVisit(ForStatement node) {
    }

    @Override
    public final void endVisit(ForStatement node) {
        endScope();
        extendEndVisit(node);
    }

    public void extendEndVisit(ForStatement node) {
    }

    @Override
    public final boolean visit(EnhancedForStatement node) {
        scope();
        node.getParameter().accept(this);
        extendVisit(node);
        return true;
    }

    protected void extendVisit(EnhancedForStatement node) {
    }

    @Override
    public final void endVisit(EnhancedForStatement node) {
        endScope();
        extendEndVisit(node);
    }

    protected void extendEndVisit(EnhancedForStatement node) {
        // see subclasses
    }

    @Override
    public boolean visit(CatchClause node) {
        scope();
        node.getException().accept(this);
        return true;
    }

    @Override
    public final void endVisit(CatchClause node) {
        endScope();
    }

    @Override
    public final boolean visit(Block node) {
        scope();
        return true;
    }

    @Override
    public final void endVisit(Block node) {
        endScope();
    }

    @Override
    public final boolean visit(MethodDeclaration node) {
        currentMethodDeclaration.push(node);
        if (CoCoUtil.isPredicate(node)) {
            return false;
        }
        scope();
        for (Object obj : node.parameters()) {
            ((SingleVariableDeclaration) obj).accept(this);
        }
        return extendVisit(node);
    }

    protected boolean extendVisit(MethodDeclaration node) {
        return true;
    }

    @Override
    public final void endVisit(MethodDeclaration node) {
        extendEndVisit(node);
        currentMethodDeclaration.pop();
        endScope();
    }

    protected void extendEndVisit(MethodDeclaration node) {
        // NOP
    }

    @Override
    public final boolean visit(SingleVariableDeclaration node) {
        variablesInScope.push(node.getName().getFullyQualifiedName());
        return true;
    }

    @Override
    public final boolean visit(VariableDeclarationFragment node) {
        if (node.getInitializer() == null && (node.resolveBinding().getJavaElement() instanceof LocalVariable)) {
            // ignore local that is not initialized
            return true;
        }
        variablesInScope.push(node.getName().getFullyQualifiedName());
        return true;
    }

    @Override
    public final boolean visit(Assignment node) {
        if (node.getLeftHandSide() instanceof SimpleName) {
            SimpleName simpleName = (SimpleName) node.getLeftHandSide();
            if (simpleName.resolveBinding().getJavaElement() instanceof LocalVariable) {
                // local variable is initialized add to scope
                // TODO: can check if it is already in scope
                variablesInScope.push(simpleName.getFullyQualifiedName());
            }
        }
        return true;
    }

    // @Override
    // public final boolean visit(CompilationUnit node) {
    // for (Object importDeclaration : node.imports()) {
    // ImportDeclaration id = (ImportDeclaration) importDeclaration;
    // if (id.getName().toString()
    // .equals(args.getFullyQualifiedExtractedPredicatesName()))
    // isImportAdded = true;
    // }
    // if (!isImportAdded) {
    // ImportDeclaration id = ast.newImportDeclaration();
    // id.setName(ast.newQualifiedName(
    // ast.newName(args.getPackageNameExtractedPredicates()),
    // ast.newSimpleName(args.getClassNameExtractedPredicates())));
    // node.imports().add(id);
    // }
    // return true;
    // }

    private boolean isPredicateAccessible(IType predicateDeclaringType, String[] parameterNames)
            throws JavaModelException {
        IType currentType = typeScope.peek();
        while (currentType != null) {
            if (currentType.getFullyQualifiedName().equals(predicateDeclaringType.getFullyQualifiedName())
                    && variablesInScope.containsAll(Arrays.asList(parameterNames))) {
                return true;
            }
            if (currentType.isMember() && !Flags.isStatic(currentType.getFlags())) {
                currentType = currentType.getDeclaringType();
            } else {
                break;
            }
        }
        return false;
    }

    protected void insertPredicate(Statement node) {
        insertPredicate(node, CoCoInstrumentationOpt.TOUCH_METHOD);
    }

    @SuppressWarnings("unchecked")
    protected void insertPredicate(Statement node, String touchMethodName) {
        try {
            List<Expression> predicateInvocations = new LinkedList<Expression>();
            Map<IMethod, IType> methodToDeclType = typePredicateMap.getPredicates(typeScope.peek());
            for (Entry<IMethod, IType> e : methodToDeclType.entrySet()) {
                IMethod predicate = e.getKey();
                IType type = e.getValue();
                if (Modifier.isStatic(currentMethodDeclaration.peek().getModifiers())
                        && !Flags.isStatic(predicate.getFlags())) {
                    continue;
                }
                String[] parameterNames = predicate.getParameterNames();
                if (isPredicateAccessible(type, parameterNames)) {
                    // Expression className = ast.newThisExpression(); //
                    // newSimpleName(args
                    // .getClassNameExtractedPredicates());
                    MethodInvocation mi = ASTBuilder.createMethodInvocation(ast(), (Expression) null,
                            predicate.getElementName(), createParameterExpressions(parameterNames));
                    // int index = whereToInsert(statements, node);
                    // statements.add(index, ast.newExpressionStatement(mi));
                    predicateInvocations.add(mi);
                }
            }
            if (predicateInvocations.size() > 0) {
                pointId++;
                MethodInvocation instInvocation = CoCoASTBuilder.createInstMethodInvocation(ast());
                // FieldAccess instAccess = ast.newFieldAccess();
                // instAccess.setName(ast.newSimpleName("inst"));
                // instAccess.setExpression(classname);
                predicateInvocations.add(0, ast().newNumberLiteral(Long.toString(pointId)));
                MethodInvocation mi = ASTBuilder.createMethodInvocation(ast(), instInvocation, touchMethodName,
                        predicateInvocations.toArray(new Expression[0]));
                @SuppressWarnings("rawtypes")
                List statements = getStatements(node);
                int index = whereToInsert(statements, node);
                statements.add(index, ast().newExpressionStatement(mi));
            }
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    protected abstract int whereToInsert(List<?> statements, Statement node);

    protected abstract List<?> getStatements(Statement node);

    protected final Expression[] createParameterExpressions(String[] parameterNames) {
        Expression[] res = new Expression[parameterNames.length];
        // res[0] = ast.newNumberLiteral(Integer.toString(pointId));
        // res[1] = Modifier.isStatic(currentMethodDeclaration.getModifiers()) ?
        // ast
        // .newNullLiteral() : ast.newThisExpression();
        for (int i = 0; i < parameterNames.length; i++) {
            res[i] = ast().newSimpleName(parameterNames[i]);
        }
        return res;
    }

    protected final void scope() {
        variablesInScope.push(SCOPE_DELIMITER);
    }

    protected final void endScope() {
        while (variablesInScope.size() > 0 && !variablesInScope.peek().equals(SCOPE_DELIMITER)) {
            variablesInScope.pop();
        }
        if (variablesInScope.size() > 0) {
            variablesInScope.pop();
        }
    }

    public final long getPointId() {
        return pointId;
    }

}
