package coco.task.instantiate;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IncrementalProjectBuilder;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jface.text.Document;
import org.eclipse.jface.text.IDocument;
import org.eclipse.text.edits.UndoEdit;

import coco.CoCoInstrumentationOpt;
import coco.log.Log;
import coco.task.EmptyProgressMonitor;
import coco.task.ITask;
import coco.task.etc.InsertMethodTouchASTV;
import coco.task.etc.RemoveIncorrectContextASTV;
import coco.task.instantiate.bc.InstantiateBC;
import coco.task.instantiate.fase.InstantiateFASE;
import coco.task.instantiate.pctbb.InstantiatePCTBB;
import coco.task.instantiate.pctst.InstantiatePCTST;
import coco.task.instantiate.sc.InstantiateSC;
import coco.util.ASTVHelper;
import coco.util.ICUCompiler;
import coco.util.ICUCompiler.ICUCompilerResult;

/**
 * Task that instantiates predicates at appropriate points. This task perform
 * the following steps:
 * 
 * <ul>
 * <li>Fix problem with [] location because of eclipse builder</li>
 * <li>Collect pure predicates</li>
 * <li>Insert method invocations to collect coverage</li>
 * <li>Insert try/finally to surround each method body</li>
 * </ul>
 * 
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 * @author Rohan Sharma (sharma.rohan1990@gmail.com)
 */
public class InstantiateTask implements ITask {

    private CoCoInstrumentationOpt args;
    private IProject project;

    private int numOfPredicates;
    private long numOfPoints;

    /** Total number of points; it is needed for continuous counting */
    private long continuousNumOfPoints;

    public InstantiateTask(CoCoInstrumentationOpt args, IProject project) {
        this.args = args;
        this.project = project;
    }

    @Override
    public void perform(ICompilationUnit icu) throws Exception {
        numOfPoints = 0;
        numOfPredicates = 0;

        // fixes problems with [] location
        MoveExtraDimensionsASTV moveExtraDimensionsASTVisitor = new MoveExtraDimensionsASTV();
        ASTVHelper.acceptVisitorRewrite(project, icu, moveExtraDimensionsASTVisitor, false);

        // remove predicates with incorrect "static context"
        // TODO: inspect if this is still needed
        ICUCompilerResult compilerResult = ICUCompiler.inst().compile(icu);
        if (!compilerResult.isCompilationSuccessful()) {
            Log.e(getClass().getName() + " " + compilerResult.getErrors());
            RemoveIncorrectContextASTV incorrectContextV = new RemoveIncorrectContextASTV(compilerResult);
            ASTVHelper.acceptVisitor(project, icu, incorrectContextV, false);
        }

        // collect side effect free predicates
        CollectTypePredicatesMappingASTV collectPurePredicatesV = new CollectTypePredicatesMappingASTV(icu);
        ASTVHelper.acceptNoModify(project, icu, collectPurePredicatesV, true);
        Log.m("Number of side effect free predicates " + icu.getElementName() + " "
                + collectPurePredicatesV.getTypePredicateMap().size());

        ITypePredicatesMap typePredicateMap = collectPurePredicatesV.getTypePredicateMap();
        InstantiateASTV instantiateASTVisitor;
        if (args.isPCTST.g()) {
            instantiateASTVisitor = new InstantiatePCTST(args, typePredicateMap, continuousNumOfPoints);
        } else if (args.isStatementCoverage.g()) {
            instantiateASTVisitor = new InstantiateSC(args, typePredicateMap, continuousNumOfPoints);
        } else if (args.isBranchCoverage.g()) {
            instantiateASTVisitor = new InstantiateBC(args, typePredicateMap, continuousNumOfPoints);
        } else if (args.isPCTMS.g()) {
            instantiateASTVisitor = new InstantiateFASE(args, typePredicateMap, continuousNumOfPoints);
        } else {
            instantiateASTVisitor = new InstantiatePCTBB(args, typePredicateMap, continuousNumOfPoints);
        }

        // insert touch methods at appropriate points
        IDocument doc = new Document(icu.getSource());
        UndoEdit undoEdit = ASTVHelper.acceptVisitor(project, icu, instantiateASTVisitor, true);

        compilerResult = ICUCompiler.inst().compile(icu);
        if (compilerResult.hasMatchingError("is exceeding the 65535 bytes limit")) {
            Log.e("Large compilation unit " + icu.getPath() + " " + icu.getElementName());
            // if the code is over limit do not apply instrumentation
            undoEdit.apply(doc);
            icu.getBuffer().setContents(doc.get());
            icu.save(EmptyProgressMonitor.inst(), true);
        } else {
            numOfPredicates += typePredicateMap.size();
            if (args.isContinuousPredicateCounting()) {
                numOfPoints = instantiateASTVisitor.getPointId() - continuousNumOfPoints;
                continuousNumOfPoints = instantiateASTVisitor.getPointId();
            } else {
                numOfPoints = instantiateASTVisitor.getPointId();
            }
        }

        // insert try/finally to surround each method body
        ASTVHelper.acceptVisitor(project, icu, new InsertMethodTouchASTV(), false);
        project.build(IncrementalProjectBuilder.FULL_BUILD, EmptyProgressMonitor.inst());
    }

    public int getNumOfPredicates() {
        return numOfPredicates;
    }

    public long getNumOfPoints() {
        return numOfPoints;
    }

    public long getContinuousNumOfPoints() {
        return continuousNumOfPoints;
    }

    public void setContinuousNumOfPoints(long n) {
        continuousNumOfPoints = n;
    }

}
