package coco.task.instantiate.pctbb;

import java.util.List;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.Block;
import org.eclipse.jdt.core.dom.CatchClause;
import org.eclipse.jdt.core.dom.DoStatement;
import org.eclipse.jdt.core.dom.EnhancedForStatement;
import org.eclipse.jdt.core.dom.ForStatement;
import org.eclipse.jdt.core.dom.IfStatement;
import org.eclipse.jdt.core.dom.LabeledStatement;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.Statement;
import org.eclipse.jdt.core.dom.SwitchCase;
import org.eclipse.jdt.core.dom.SwitchStatement;
import org.eclipse.jdt.core.dom.TryStatement;
import org.eclipse.jdt.core.dom.WhileStatement;

import coco.CoCoInstrumentationOpt;
import coco.task.instantiate.ITypePredicatesMap;
import coco.task.instantiate.InsertBasicBlockASTV;
import coco.task.instantiate.InstantiateASTV;
import coco.util.CoCoUtil;

/**
 * Code instrumentation for measuring PCT(BB). Instantiates predicates at each
 * basic block. This visitor requires {@link InsertBasicBlockASTV}.
 * 
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 * @author Rohan Sharma (sharma.rohan1990@gmail.com)
 */
public class InstantiatePCTBB extends InstantiateASTV {

    public InstantiatePCTBB(CoCoInstrumentationOpt args, ITypePredicatesMap typePredicateMap, long pointId) {
        super(args, typePredicateMap, pointId);
    }

    // ASTVisitor interface

    @Override
    public boolean extendVisit(MethodDeclaration node) {
        if (CoCoUtil.isCondition(node) || node.getBody() == null) {
            return false;
        }
        // skip "super" and "this" constructor invocations
        Block body = node.getBody();
        if (node.isConstructor()) {
            for (Object o : body.statements()) {
                if (o instanceof Block) {
                    body = (Block) o;
                    break;
                }
            }
        }
        insertPredicate(body, CoCoInstrumentationOpt.TOUCH_START_METHOD);
        return true;
    }

    @Override
    public boolean visit(IfStatement node) {
        insertPredicate(node.getThenStatement(), CoCoInstrumentationOpt.TOUCH_BRANCH_METHOD);
        insertPredicate(node.getElseStatement(), CoCoInstrumentationOpt.TOUCH_BRANCH_METHOD);
        insertPredicatesAfter(node, CoCoInstrumentationOpt.TOUCH_METHOD);
        return true;
    }

    @Override
    public void extendVisit(ForStatement node) {
        insertPredicate(node.getBody(), CoCoInstrumentationOpt.TOUCH_BRANCH_METHOD);
    }

    @Override
    public void extendEndVisit(ForStatement node) {
        insertPredicatesAfter(node, CoCoInstrumentationOpt.TOUCH_BRANCH_METHOD);
    }

    @Override
    public void extendVisit(EnhancedForStatement node) {
        insertPredicate(node.getBody(), CoCoInstrumentationOpt.TOUCH_BRANCH_METHOD);
    }

    @Override
    public void extendEndVisit(EnhancedForStatement node) {
        insertPredicatesAfter(node, CoCoInstrumentationOpt.TOUCH_BRANCH_METHOD);
    }

    @Override
    public boolean visit(WhileStatement node) {
        insertPredicate(node.getBody(), CoCoInstrumentationOpt.TOUCH_BRANCH_METHOD);
        insertPredicatesAfter(node, CoCoInstrumentationOpt.TOUCH_BRANCH_METHOD);
        return true;
    }

    @Override
    public boolean visit(DoStatement node) {
        insertPredicate(node.getBody(), CoCoInstrumentationOpt.TOUCH_BRANCH_METHOD);
        insertPredicatesAfter(node, CoCoInstrumentationOpt.TOUCH_BRANCH_METHOD);
        return true;
    }

    @Override
    public boolean visit(CatchClause node) {
        if (!super.visit(node)) {
            return false;
        }
        insertPredicate(node.getBody(), CoCoInstrumentationOpt.TOUCH_METHOD);
        return true;
    }

    @Override
    public boolean visit(TryStatement node) {
        insertPredicate(node.getBody(), CoCoInstrumentationOpt.TOUCH_METHOD);
        insertPredicate(node.getFinally(), CoCoInstrumentationOpt.TOUCH_METHOD);
        insertPredicatesAfter(node, CoCoInstrumentationOpt.TOUCH_METHOD);
        return true;
    }

    @Override
    public boolean visit(SwitchStatement node) {
        for (int i = 0; i < node.statements().size(); i++) {
            Statement s = (Statement) node.statements().get(i);
            if (!(s instanceof SwitchCase)) {
                insertPredicate(s, CoCoInstrumentationOpt.TOUCH_BRANCH_METHOD);
            }
        }
        insertPredicatesAfter(node, CoCoInstrumentationOpt.TOUCH_METHOD);
        return true;
    }

    // PROTECTED

    @Override
    protected List<?> getStatements(Statement node) {
        return ((Block) node).statements();
    }

    @Override
    protected int whereToInsert(List<?> statements, Statement node) {
        return 0;
    }

    // PRIVATE

    private void insertPredicatesAfter(ASTNode node, String touchMethodName) {
        Block block;
        if (node.getParent() instanceof LabeledStatement) {
            block = (Block) node.getParent().getParent();
            // node has to be faked! (see the use of node below)
            node = node.getParent();
        } else {
            block = (Block) node.getParent();
        }
        int index = block.statements().indexOf(node);
        if (index + 1 < block.statements().size()) {
            // maybe we have removed after block (unreachable code)
            Block after = (Block) block.statements().get(index + 1);
            insertPredicate(after, touchMethodName);
        }
    }
}
