package coco.task.renamevars;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.util.HashSet;
import java.util.Set;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IField;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.ILocalVariable;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.internal.corext.refactoring.rename.JavaRenameProcessor;
import org.eclipse.jdt.internal.corext.refactoring.rename.RenameFieldProcessor;
import org.eclipse.jdt.internal.corext.refactoring.rename.RenameLocalVariableProcessor;
import org.eclipse.jdt.internal.corext.refactoring.util.TextChangeManager;
import org.eclipse.ltk.core.refactoring.GroupCategorySet;
import org.eclipse.ltk.core.refactoring.Refactoring;
import org.eclipse.ltk.core.refactoring.participants.RenameRefactoring;

import coco.task.EmptyProgressMonitor;
import coco.task.ITask;
import coco.task.PerformRefactoring;
import coco.task.ProblemReport;

/**
 * Renames variables to avoid duplicate names before instrumentation.
 * 
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 * @author Rohan Sharma (sharma.rohan1990@gmail.com)
 */
@SuppressWarnings("restriction")
public class RenameVarsTask implements ITask {

    private static final String NEW_NAME_SUFFIX = "__";

    /** Project that contains code to instrument */
    private final IProject project;

    /** Names that have been seen */
    private final Set<String> names;

    /** Unchanged names because there is an error reported by refactoring */
    private final Set<String> conflicts;

    public RenameVarsTask(IProject project) {
        this.project = project;
        this.names = new HashSet<String>();
        this.conflicts = new HashSet<String>();
    }

    @Override
    public void perform(ICompilationUnit icu) throws Exception {
        RenameVarsASTV renameASTVisitor = new RenameVarsASTV(names);
        while (true) {
            renameASTVisitor.advanceId();
            ASTParser parser = ASTParser.newParser(AST.JLS3);
            parser.setResolveBindings(true);
            parser.setSource(icu);
            ASTNode root = parser.createAST(EmptyProgressMonitor.inst());
            root.accept(renameASTVisitor);
            if (!renameASTVisitor.hasVariable()) {
                break;
            }
            Refactoring refactoring = null;
            IJavaElement iJavaElement = renameASTVisitor.getElement();
            String newName = iJavaElement.getElementName() + NEW_NAME_SUFFIX + renameASTVisitor.getId();
            if (iJavaElement instanceof IField) {
                refactoring = getFieldRefactoring((IField) iJavaElement, newName);
            } else if (renameASTVisitor.getElement() instanceof ILocalVariable) {
                refactoring = getLocalRefactoring((ILocalVariable) renameASTVisitor.getElement(), renameASTVisitor,
                        newName);
            } else {
                throw new RuntimeException();
            }
            ProblemReport report = PerformRefactoring.performRefactoring(refactoring, true);
            if (report.hasError()) {
                PerformRefactoring.undoRefactoring(PerformRefactoring.getUndoChange());
                conflicts.add(iJavaElement.getElementName());
            }
            project.refreshLocal(IResource.DEPTH_INFINITE, EmptyProgressMonitor.inst());
        }
    }

    // PRIVATE

    private Refactoring getRefactoring(JavaRenameProcessor processor, String newName) throws Exception {
        processor.setNewElementName(newName);
        RenameRefactoring refactoring = new RenameRefactoring(processor);
        refactoring.checkInitialConditions(EmptyProgressMonitor.inst());
        Field fIsComposite = processor.getClass().getDeclaredField("fIsComposite");
        fIsComposite.setAccessible(true);
        fIsComposite.setBoolean(processor, false);
        return refactoring;
    }

    private Refactoring getFieldRefactoring(IField field, String newName) throws Exception {
        RenameFieldProcessor processor = new RenameFieldProcessor(field);
        return getRefactoring(processor, newName);
    }

    private Refactoring getLocalRefactoring(ILocalVariable local, RenameVarsASTV renameASTVisitor, String newName)
            throws Exception {
        Class<?> clz = RenameLocalVariableProcessor.class;
        Constructor<?> constructor = clz.getDeclaredConstructor(ILocalVariable.class, TextChangeManager.class,
                CompilationUnit.class, GroupCategorySet.class);
        constructor.setAccessible(true);
        RenameLocalVariableProcessor processor = (RenameLocalVariableProcessor) constructor.newInstance(local,
                new TextChangeManager(true), renameASTVisitor.getCompilationUnit(), GroupCategorySet.NONE);
        processor.setUpdateReferences(true);
        Field fTempDeclarationNode = processor.getClass().getDeclaredField("fTempDeclarationNode");
        fTempDeclarationNode.setAccessible(true);
        fTempDeclarationNode.set(processor, renameASTVisitor.getParentVariableDeclaration());
        // Field fIsComposite =
        // processor.getClass().getDeclaredField("fIsComposite");
        // fIsComposite.setAccessible(true);
        // fIsComposite.setBoolean(processor, true);
        // Field fCompilationUnitNode = processor.getClass().getDeclaredField(
        // "fCompilationUnitNode");
        // fCompilationUnitNode.setAccessible(true);
        // fCompilationUnitNode.set(processor, ast.getCompilationUnit());
        return getRefactoring(processor, newName);
    }

}
