package coco.util;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ArrayAccess;
import org.eclipse.jdt.core.dom.Block;
import org.eclipse.jdt.core.dom.CastExpression;
import org.eclipse.jdt.core.dom.CatchClause;
import org.eclipse.jdt.core.dom.CharacterLiteral;
import org.eclipse.jdt.core.dom.ClassInstanceCreation;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.FieldAccess;
import org.eclipse.jdt.core.dom.IfStatement;
import org.eclipse.jdt.core.dom.ImportDeclaration;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.Name;
import org.eclipse.jdt.core.dom.PrimitiveType;
import org.eclipse.jdt.core.dom.ReturnStatement;
import org.eclipse.jdt.core.dom.SimpleType;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.Statement;
import org.eclipse.jdt.core.dom.TryStatement;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.jdt.core.dom.TypeLiteral;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;
import org.eclipse.jdt.core.dom.Modifier.ModifierKeyword;
import org.eclipse.jdt.core.dom.PrimitiveType.Code;

/**
 * Utility class that provides helper method for building commonly used
 * {@link ASTNode}s.
 * 
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 * @author Rohan Sharma (sharma.rohan1990@gmail.com)
 */
@SuppressWarnings("unchecked")
public class ASTBuilder {

    private static final String MAIN_LITERAL = "main";
    private static final String PRINTLN_LITERAL = "println";

    private static final String STRING_LITERAL = "String";
    private static final String SYSTEM_LITERAL = "System";
    private static final String EXCEPTION_LITERAL = "Exception";

    private static final String ARGS_VAR_LITERAL = "args";
    private static final String OUT_VAR_LITERAL = "out";
    private static final String EX_VAR_LITERAL = "__ex";

    /**
     * Builds an returns an {@link Expression} that has a value for the given
     * {@link Type}.
     * 
     * @param type
     * @return an {@link Expression} that has a value for the given {@link Type}
     */
    public static Expression getValueLiteralForType(AST ast, Type type) {
        Expression valueLiteral = ast.newNullLiteral();

        if (type.isPrimitiveType()) {

            PrimitiveType primType = (PrimitiveType) type;
            Code primTypeCode = primType.getPrimitiveTypeCode();

            if (primTypeCode.equals(PrimitiveType.BOOLEAN)) {
                valueLiteral = ast.newBooleanLiteral(false);
            } else if (primTypeCode.equals(PrimitiveType.BYTE) || primTypeCode.equals(PrimitiveType.DOUBLE)
                    || primTypeCode.equals(PrimitiveType.FLOAT) || primTypeCode.equals(PrimitiveType.INT)
                    || primTypeCode.equals(PrimitiveType.LONG) || primTypeCode.equals(PrimitiveType.SHORT)) {
                valueLiteral = ast.newNumberLiteral();
            } else if (primTypeCode.equals(PrimitiveType.CHAR)) {
                valueLiteral = ast.newCharacterLiteral();
                ((CharacterLiteral) valueLiteral).setCharValue('c');
            }

        }

        return valueLiteral;
    }

    /**
     * Builds and returns a {@link List} of {@link Expression}s that have values
     * for the given list of {@link Type}s.
     * 
     * @param types
     * @return a {@link List} of {@link Expression}s that have values for the
     *         given list of {@link Type}s.
     */
    public static List<Expression> getValueLiteralsForTypes(AST ast, Type... types) {
        ArrayList<Expression> valueLiterals = new ArrayList<Expression>();

        for (Type type : types) {
            valueLiterals.add(getValueLiteralForType(ast, type));
        }

        return valueLiterals;
    }

    public static ImportDeclaration createImport(AST ast, boolean isOnDemand, boolean isStatic, String... pck) {
        ImportDeclaration importDeclaration = ast.newImportDeclaration();
        Name name = ast.newName(pck);
        importDeclaration.setName(name);
        importDeclaration.setOnDemand(isOnDemand);
        importDeclaration.setStatic(isStatic);
        return importDeclaration;
    }

    public static TypeDeclaration createTypeDeclaration(AST ast, boolean isPublic, boolean isInterface, String name) {
        TypeDeclaration typeDeclaration = ast.newTypeDeclaration();
        typeDeclaration.setInterface(isInterface);
        typeDeclaration.setName(ast.newSimpleName(name));
        if (isPublic)
            typeDeclaration.modifiers().add(ASTBuilder.createPublicModifier(ast));
        return typeDeclaration;
    }

    public static MethodDeclaration createMainMethodDeclaration(AST ast) {
        MethodDeclaration methodDeclaration = ast.newMethodDeclaration();
        methodDeclaration.setName(ast.newSimpleName(MAIN_LITERAL));
        methodDeclaration.modifiers().add(ast.newModifier(ModifierKeyword.PUBLIC_KEYWORD));
        methodDeclaration.modifiers().add(ast.newModifier(ModifierKeyword.STATIC_KEYWORD));
        methodDeclaration.setReturnType2(ast.newPrimitiveType(PrimitiveType.VOID));
        SingleVariableDeclaration mainArgument = ast.newSingleVariableDeclaration();
        mainArgument.setName(ast.newSimpleName(ARGS_VAR_LITERAL));
        mainArgument.setType(ast.newArrayType(ast.newSimpleType(ast.newName(STRING_LITERAL))));
        methodDeclaration.parameters().add(mainArgument);
        methodDeclaration.setBody(ast.newBlock());
        return methodDeclaration;
    }

    public static MethodDeclaration createMethodDeclaration(AST ast, String name, Number returnValue) {
        MethodDeclaration methodDeclaration = ast.newMethodDeclaration();
        methodDeclaration.setName(ast.newSimpleName(name));
        methodDeclaration.modifiers().add(ast.newModifier(ModifierKeyword.PUBLIC_KEYWORD));
        // methodDeclaration.modifiers().add(
        // ast.newModifier(ModifierKeyword.STATIC_KEYWORD));
        methodDeclaration.setReturnType2(ast.newPrimitiveType(PrimitiveType.INT));
        CastExpression castExpression = ast.newCastExpression();
        castExpression.setExpression(ast.newNumberLiteral(returnValue.toString()));
        castExpression.setType(ast.newPrimitiveType(PrimitiveType.INT));
        ReturnStatement returnStatement = ast.newReturnStatement();
        returnStatement.setExpression(castExpression);
        Block body = ast.newBlock();
        body.statements().add(returnStatement);
        methodDeclaration.setBody(body);
        return methodDeclaration;
    }

    public static MethodInvocation createSysout(AST ast, Expression expression) {
        FieldAccess sysout = ASTBuilder.createFieldAccess(ast, SYSTEM_LITERAL, OUT_VAR_LITERAL);
        MethodInvocation println = ASTBuilder.createMethodInvocation(ast, sysout, PRINTLN_LITERAL, expression);
        return println;
    }

    public static MethodInvocation createSysout(AST ast, String expression) {
        return createSysout(ast, ast.newSimpleName(expression));
    }

    public static TryStatement createTryCatchStatement(AST ast) {
        SingleVariableDeclaration exception = ast.newSingleVariableDeclaration();
        exception.setType(ast.newSimpleType(ast.newName(EXCEPTION_LITERAL)));
        exception.setName(ast.newSimpleName(EX_VAR_LITERAL));

        CatchClause catchClause = ast.newCatchClause();
        catchClause.setBody(ast.newBlock());
        catchClause.setException(exception);
        catchClause.getBody().statements().add(ast.newExpressionStatement(createSysout(ast, EX_VAR_LITERAL)));

        TryStatement tryStatement = ast.newTryStatement();
        tryStatement.setBody(ast.newBlock());
        tryStatement.catchClauses().add(catchClause);
        return tryStatement;
    }

    public static TryStatement createTryFinally(AST ast, List<Statement> tryStatements,
            List<Statement> finallyStatements) {
        Block tryBlock = ast.newBlock();
        tryBlock.statements().addAll(tryStatements);
        Block finallyBlock = ast.newBlock();
        finallyBlock.statements().addAll(finallyStatements);
        return createTryFinally(ast, tryBlock, finallyBlock);
    }

    public static TryStatement createTryFinally(AST ast, Block tryBlock, Block finallyBlock) {
        TryStatement tryStatement = ast.newTryStatement();
        tryStatement.setBody(tryBlock);
        tryStatement.setFinally(finallyBlock);
        return tryStatement;
    }

    public static VariableDeclarationStatement createVariableDeclarationStatement(AST ast, String typeOfFields,
            VariableDeclarationFragment... fragments) {
        return ASTBuilder.createVariableDeclarationStatement(ast, ast.newSimpleType(ast.newName(typeOfFields)),
                fragments);
    }

    public static VariableDeclarationStatement createVariableDeclarationStatement(AST ast, Type typeOfFields,
            VariableDeclarationFragment... fragments) {
        VariableDeclarationStatement variableDeclarationStatement = ast.newVariableDeclarationStatement(fragments[0]);

        variableDeclarationStatement.setType(typeOfFields);
        for (int i = 1; i < fragments.length; i++)
            variableDeclarationStatement.fragments().add(fragments[i]);
        return variableDeclarationStatement;
    }

    public static VariableDeclarationFragment createVariableDeclarationFragment(AST ast, String fieldName,
            Expression initializer) {
        VariableDeclarationFragment vdf = ast.newVariableDeclarationFragment();
        vdf.setName(ast.newSimpleName(fieldName));
        vdf.setInitializer(initializer);
        return vdf;
    }

    public static MethodInvocation createMethodInvocation(AST ast, Expression expression, String methodName) {
        MethodInvocation methodInvocation = ast.newMethodInvocation();
        methodInvocation.setName(ast.newSimpleName(methodName));
        methodInvocation.setExpression(expression);
        return methodInvocation;
    }

    public static MethodInvocation createMethodInvocation(AST ast, Expression expression, String methodName,
            Expression... arguments) {
        MethodInvocation methodInvocation = ast.newMethodInvocation();
        methodInvocation.setName(ast.newSimpleName(methodName));
        methodInvocation.setExpression(expression);
        for (int i = 0; i < arguments.length; i++)
            methodInvocation.arguments().add(arguments[i]);
        return methodInvocation;
    }

    public static MethodInvocation createMethodInvocation(AST ast, String fieldName, String methodName,
            Expression... arguments) {
        return createMethodInvocation(ast, ast.newSimpleName(fieldName), methodName, arguments);
    }

    public static MethodInvocation createMethodInvocationStringArgs(AST ast, String fieldName, String methodName,
            String... arguments) {
        return ASTBuilder.createMethodInvocationStringArgs(ast, ast.newSimpleName(fieldName), methodName, arguments);
    }

    public static MethodInvocation createMethodInvocationStringArgs(AST ast, Expression expression, String methodName,
            String... arguments) {
        Expression[] expressions = new Expression[arguments.length];
        for (int i = 0; i < arguments.length; i++) {
            expressions[i] = ast.newSimpleName(arguments[i]);
        }
        return createMethodInvocation(ast, expression, methodName, expressions);
    }

    public static FieldAccess createFieldAccess(AST ast, Expression expression, String fieldName) {
        FieldAccess fa = ast.newFieldAccess();
        fa.setExpression(expression);
        fa.setName(ast.newSimpleName(fieldName));
        return fa;
    }

    public static FieldAccess createFieldAccess(AST ast, String object, String fieldName) {
        return createFieldAccess(ast, ast.newSimpleName(object), fieldName);
    }

    public static ClassInstanceCreation createClassInstanceCreation(AST ast, Type type) {
        ClassInstanceCreation cic = ast.newClassInstanceCreation();
        cic.setType(type);
        return cic;
    }

    public static ClassInstanceCreation createClassInstanceCreation(AST ast, String type) {
        return createClassInstanceCreation(ast, ast.newSimpleType(ast.newName(type)));
    }

    public static ArrayAccess createArrayAccess(AST ast, Expression array, Expression index) {
        ArrayAccess aa = ast.newArrayAccess();
        aa.setArray(array);
        aa.setIndex(index);
        return aa;
    }

    public static ArrayAccess createArrayAccess(AST ast, Expression array, String index) {
        return createArrayAccess(ast, array, ast.newSimpleName(index));
    }

    public static Modifier createPublicModifier(AST ast) {
        return ast.newModifier(ModifierKeyword.PUBLIC_KEYWORD);
    }

    public static TypeLiteral createTypeLiteral(AST ast, String className) {
        TypeLiteral tl = ast.newTypeLiteral();
        tl.setType(ast.newSimpleType(ast.newName(className)));
        return tl;
    }

    public static SimpleType createSimpleType(AST ast, String type) {
        return ast.newSimpleType(ast.newName(type));
    }

    public static IfStatement createIfThenStatement(AST ast, Expression expression, Statement thenStatement) {
        IfStatement ifStatement = ast.newIfStatement();
        ifStatement.setExpression(expression);
        ifStatement.setThenStatement(thenStatement);
        return ifStatement;
    }

}
