package coco.util;

import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.Block;
import org.eclipse.jdt.core.dom.CatchClause;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.IfStatement;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.ReturnStatement;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.TryStatement;

import coco.CoCoInstrumentationOpt;

/**
 * Builds on top of {@link ASTBuilder} to provide methods specific to CoCo.
 * 
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 * @author Rohan Sharma (sharma.rohan1990@gmail.com)
 */
public class CoCoASTBuilder {

    /** Variable name for catch block */
    private static final String EXCEPTION_VAR = "_coco_ex";

    public static MethodInvocation createInstMethodInvocation(AST ast) {
        MethodInvocation instInvocation = ast.newMethodInvocation();
        instInvocation.setExpression(createCoCoRuntimeClassName(ast));
        instInvocation.setName(ast.newSimpleName(CoCoInstrumentationOpt.INST_METHOD));
        return instInvocation;
    }

    @SuppressWarnings("unchecked")
    public static TryStatement createTryFinallyIfInPredicateReturn(AST ast, Block tryBlock, String predicateId) {
        Expression testAndSet = createSetInPredicateMethodInvocation(ast);

        Block thenBlock = ast.newBlock();
        Expression ignore = createIgnoreNextTouchMethodInvocation(ast);
        thenBlock.statements().add(ast.newExpressionStatement(ignore));
        ReturnStatement returnStatement = ast.newReturnStatement();
        returnStatement.setExpression(ast.newNumberLiteral(predicateId));
        thenBlock.statements().add(returnStatement);

        IfStatement ifStatement = ASTBuilder.createIfThenStatement(ast, testAndSet, thenBlock);
        tryBlock.statements().add(0, ifStatement);

        Block finallyBlock = ast.newBlock();
        Expression reset = createResetInPredicateMethodInvocation(ast);
        finallyBlock.statements().add(ast.newExpressionStatement(reset));

        ReturnStatement returnCatch = ast.newReturnStatement();
        returnCatch.setExpression(ast.newNumberLiteral(predicateId));
        Block catchBody = ast.newBlock();
        catchBody.statements().add(returnCatch);

        SingleVariableDeclaration ex = ast.newSingleVariableDeclaration();
        ex.setName(ast.newSimpleName(EXCEPTION_VAR));
        ex.setType(ast.newSimpleType(ast.newName(Exception.class.getName())));

        CatchClause catchClause = ast.newCatchClause();
        catchClause.setException(ex);
        catchClause.setBody(catchBody);

        TryStatement tryStatement = ASTBuilder.createTryFinally(ast, tryBlock, finallyBlock);
        tryStatement.catchClauses().add(catchClause);

        return tryStatement;
    }

    // PRIVATE

    private static Expression createCoCoRuntimeClassName(AST ast) {
        Expression className = ast.newQualifiedName(ast.newName(CoCoInstrumentationOpt.COCO_RUNTIME_PCK),
                ast.newSimpleName(CoCoInstrumentationOpt.COCO_RUNTIME_CLASS));
        return className;
    }

    private static MethodInvocation createSetInPredicateMethodInvocation(AST ast) {
        MethodInvocation setInPredicateInvocation = ast.newMethodInvocation();
        setInPredicateInvocation.setExpression(createInstMethodInvocation(ast));
        setInPredicateInvocation.setName(ast.newSimpleName(CoCoInstrumentationOpt.SET_IN_PREDICATE_METHOD));
        return setInPredicateInvocation;
    }

    private static MethodInvocation createResetInPredicateMethodInvocation(AST ast) {
        MethodInvocation resetInPredicateInvocation = ast.newMethodInvocation();
        resetInPredicateInvocation.setExpression(createInstMethodInvocation(ast));
        resetInPredicateInvocation.setName(ast.newSimpleName(CoCoInstrumentationOpt.RESET_IN_PREDICATE_METHOD));
        return resetInPredicateInvocation;
    }

    private static MethodInvocation createIgnoreNextTouchMethodInvocation(AST ast) {
        MethodInvocation ignoreNextTouchInvocation = ast.newMethodInvocation();
        ignoreNextTouchInvocation.setExpression(createInstMethodInvocation(ast));
        ignoreNextTouchInvocation.setName(ast.newSimpleName(CoCoInstrumentationOpt.IGNORE_TOUCH_METHOD));
        return ignoreNextTouchInvocation;
    }

}
