package coco.util;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.IPackageFragmentRoot;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.AnonymousClassDeclaration;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.ExpressionStatement;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.SimpleName;

/**
 * Collection of util methods to analyze a part of Java project.
 * 
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 * @author Rohan Sharma (sharma.rohan1990@gmail.com)
 */
public class CoCoUtil {

    private static final String CONDITION_EXT = "$$";
    private static final String PREDICATE_EXT = "$$";

    private static final String CONDITION_PREFIX = "condition" + CONDITION_EXT;
    private static final String PREDICATE_PREFIX = "predicate" + PREDICATE_EXT;

    public static String makeConditionString(int conditionId) {
        return CONDITION_PREFIX + conditionId;
    }

    public static boolean isCondition(SimpleName node) {
        return node.getFullyQualifiedName().startsWith(CONDITION_PREFIX);
    }

    public static boolean isCondition(IMethod method) {
        return method.getElementName().startsWith(CONDITION_PREFIX);
    }

    public static boolean isCondition(org.eclipse.jdt.core.dom.MethodDeclaration methodDeclaration) {
        return methodDeclaration.getName().toString().startsWith(CONDITION_PREFIX);
    }

    // @SuppressWarnings("restriction")
    // public static boolean isCondition(
    // org.eclipse.jdt.internal.compiler.ast.MethodDeclaration
    // methodDeclaration)
    // {
    // return new
    // String(methodDeclaration.selector).startsWith(CONDITION_PREFIX);
    // }

    public static String getConditionId(MethodDeclaration condition) {
        String conditionName = condition.getName().getFullyQualifiedName();
        return conditionName.substring(conditionName.lastIndexOf(CONDITION_EXT) + CONDITION_EXT.length());
    }

    public static void findConditionsInICU(ICompilationUnit icu, List<IMethod> conditions, List<IType> types)
            throws JavaModelException {
        for (IType type : icu.getTypes()) {
            for (IMethod method : type.getMethods()) {
                if (isCondition(method)) {
                    conditions.add(method);
                    types.add(type);
                }
            }
        }
    }

    // TODO: remove this method. Deprecated until we solve todo.
    @Deprecated
    public static boolean isConditionInvocation(ExpressionStatement node) {
        return isConditionInvocation(node.getExpression());
    }

    // TODO: check when it is used. Deprecated until we solve todo.
    @Deprecated
    public static boolean isConditionInvocation(Expression node) {
        if (node instanceof MethodInvocation) {
            MethodInvocation mi = (MethodInvocation) node;
            if (mi.getName().toString().startsWith(CONDITION_PREFIX)) {
                return true;
            }
        }
        return false;
    }

    public static String makePredicateString(String predicateId) {
        return PREDICATE_PREFIX + predicateId;
    }

    public static boolean isPredicate(org.eclipse.jdt.core.dom.MethodDeclaration node) {
        return node.getName().toString().startsWith(PREDICATE_PREFIX);
    }

    public static boolean isPredicateInvocation(Expression node) {
        if (node instanceof MethodInvocation) {
            MethodInvocation mi = (MethodInvocation) node;
            if (mi.getName().toString().startsWith(PREDICATE_PREFIX)) {
                return true;
            }
        }
        return false;
    }

    public static String getPredicateId(MethodDeclaration predicate) {
        String predicateName = predicate.getName().getFullyQualifiedName();
        return predicateName.substring(predicateName.lastIndexOf(PREDICATE_EXT) + PREDICATE_EXT.length());
    }

    public static IType getDeclaringIType(MethodDeclaration methodDeclaration) {
        IMethodBinding binding = methodDeclaration.resolveBinding();
        IJavaElement javaElement = binding.getDeclaringClass().getJavaElement();
        IType iType = (IType) javaElement;
        return iType;
    }

    public static IMethod getIMethod(MethodDeclaration methodDeclaration) {
        IMethodBinding binding = methodDeclaration.resolveBinding();
        IJavaElement javaElement = binding.getJavaElement();
        IMethod iMethod = (IMethod) javaElement;
        return iMethod;
    }

    public static ICompilationUnit findACompilationUnitInProject(String name, IProject project) throws CoreException {
        if (project.isNatureEnabled("org.eclipse.jdt.core.javanature")) {
            IJavaProject javaProject = JavaCore.create(project);
            IPackageFragment[] packages = javaProject.getPackageFragments();
            for (IPackageFragment pck : packages) {
                if (pck.getKind() == IPackageFragmentRoot.K_SOURCE) {
                    for (ICompilationUnit icu : pck.getCompilationUnits()) {
                        if (icu.getElementName().toString().equals(name + ".java"))
                            return icu;
                    }
                }
            }
        }
        return null;
    }

    public static List<ICompilationUnit> findCompilationUnitsInProject(IProject project) throws CoreException {
        Comparator<ICompilationUnit> comparator = new Comparator<ICompilationUnit>() {
            @Override
            public int compare(ICompilationUnit o1, ICompilationUnit o2) {
                return o1.getElementName().compareTo(o2.getElementName());
            }
        };
        List<ICompilationUnit> icus = new LinkedList<ICompilationUnit>();
        if (project.isNatureEnabled("org.eclipse.jdt.core.javanature")) {
            IJavaProject javaProject = JavaCore.create(project);
            IPackageFragment[] packages = javaProject.getPackageFragments();
            for (IPackageFragment pck : packages) {
                if (pck.getKind() == IPackageFragmentRoot.K_SOURCE) {
                    ICompilationUnit[] iCompilationUnits = pck.getCompilationUnits();
                    Arrays.sort(iCompilationUnits, comparator);
                    for (ICompilationUnit icu : iCompilationUnits) {
                        icus.add(icu);
                    }
                }
            }
        }
        return icus;
    }

    public static MethodDeclaration[] getMethods(AnonymousClassDeclaration node) {
        List<MethodDeclaration> methods = new ArrayList<MethodDeclaration>();
        for (Object obj : node.bodyDeclarations()) {
            if (obj instanceof MethodDeclaration) {
                methods.add((MethodDeclaration) obj);
            }
        }
        return methods.toArray(new MethodDeclaration[0]);
    }

    public static String getSignature(IMethodBinding methodBinding) {
        return methodBinding.getKey();
        // String s = methodBinding.getName();
        // try {
        // Field bindingField =
        // methodBinding.getClass().getDeclaredField("binding");
        // bindingField.setAccessible(true);
        // Object binding = bindingField.get(methodBinding);
        // Class<?> bindingClass = binding.getClass();
        // Field signatureField = null;
        // while (bindingClass != null) {
        // try {
        // signatureField = bindingClass.getDeclaredField("signature");
        // } catch (NoSuchFieldException ex) {
        // bindingClass = bindingClass.getSuperclass();
        // }
        // if (signatureField != null) {
        // break;
        // }
        // }
        // assert (signatureField != null);
        // signatureField.setAccessible(true);
        // s += new String((char[]) signatureField.get(binding));
        // } catch (SecurityException e) {
        // e.printStackTrace();
        // } catch (NoSuchFieldException e) {
        // e.printStackTrace();
        // } catch (IllegalArgumentException e) {
        // e.printStackTrace();
        // } catch (IllegalAccessException e) {
        // e.printStackTrace();
        // }
        // return s;
    }

    public static String getSignature(MethodDeclaration node) {
        IMethodBinding methodBinding = node.resolveBinding();
        return getSignature(methodBinding);
    }
}
