package coring;

/**
 * Describes a command line argument. Fields of this type must be
 * <code>final</code>.
 * 
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 */
abstract class Opt<T> {
    private static final String SHORT_INDENT = "  ";
    private static final String INDENT = "    ";

    @SuppressWarnings("unused")
    private final T defaultValue;
    private T value;

    /** name of the option */
    private final String name;
    /** short name of the option */
    private final String shortName;

    /** help message for the option */
    private String helpMsg;

    /** is option required on command line */
    private boolean isRequired;

    /** is option provided on command line */
    private boolean isProvided;

    public Opt(String name, Class<?> clz) {
        this(name, null, clz, null);
    }

    public Opt(String name, Class<?> clz, T value) {
        this(name, null, clz, value);
    }

    public Opt(String name, String shortName, Class<?> clz) {
        this(name, shortName, clz, null);
    }

    public Opt(String name, String shortName, Class<?> clz, T value) {
        if (clz == null || name == null) {
            throw new NullPointerException();
        }
        this.name = OptConfig.LONG_ARG_PREFIX + name;
        this.shortName = shortName == null ? null : OptConfig.SHORT_ARG_PREFIX + shortName;
        this.value = value;
        this.defaultValue = value;
    }

    public boolean isProvided() {
        return isProvided;
    }

    public T g() {
        // may want to clone
        return value;
    }

    protected abstract void setVal(String[] args, int index);

    public abstract Opt<T> s(T v);

    /**
     * Set value of the command line option. It should be used carefully.
     */
    protected void unsafeSet(T v) {
        value = v;
    }

    @Override
    public String toString() {
        return (shortName == null ? "" : (shortName + ", ")) + name;
    }

    // internal methods

    /**
     * Parses/saves the value from the given arguments. This method is invoked
     * when name of this object is the same as the option provided at position
     * index.
     */
    protected final void parseValue(String[] args, int index) {
        isProvided = true;
        setVal(args, index);
    }

    /**
     * Returns string to describe type of the argument for an option.
     */
    protected abstract String getArgArgumentString();

    String getName() {
        return name;
    }

    String getShortName() {
        return shortName;
    }

    String getHelp() {
        // consider adding default value
        StringBuilder sb = new StringBuilder(INDENT);
        if (shortName != null) {
            sb.append(shortName).append(", ");
        }
        sb.append(name);
        if (isRequired) {
            sb.append(OptConfig.REQUIRED_HELP);
        }
        if (getArgArgumentString() != null) {
            sb.append(SHORT_INDENT).append(getArgArgumentString());
        }
        if (helpMsg != null) {
            sb.append("\n").append(INDENT).append(INDENT).append(helpMsg);
        }
        return sb.toString();
    }

    void setHelpMsg(String s) {
        helpMsg = s;
    }

    boolean isRequired() {
        return isRequired;
    }

    void setRequired(boolean f) {
        isRequired = f;
    }
}
