package coring;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

/**
 * Parser for command line arguments. Any object can be used to keep arguments
 * (fields of type {@link Opt}). {@link Opt}s are discovered automatically and
 * appropriate values are assigned.
 * 
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 */
public final class OptParser {
    /** Help opt */
    private static final BooleanOpt HELP_ARG = new BooleanOpt("help", "h");

    /** Config file opt */
    private static final StringOpt CONFIG_ARG = new StringOpt("config", "c");

    static {
        HELP_ARG.setHelpMsg("Print this help message");
        CONFIG_ARG.setHelpMsg("File with option value pairs");
    }

    public static OptParserStatus parse(Object options, String[] args) {
        if (options == null || args == null) {
            throw new NullPointerException();
        }
        OptParserStatus status = new OptParserStatus();
        try {
            List<Field> argFields = findArgFields(options);
            checkArgsDef(options, argFields);
            // load options from configuration file (if provided)
            args = loadConfig(args, configFileName(args));
            checkCmd(options, args, argFields);
            assignAnnotations(options, argFields);
            assignValues(options, args, argFields);
            // checkArgsValues(options, argFields);
            executeArgAfter(options);
            status.setSuccessfullyParsed(true);
        } catch (OptException e) {
            // e.printStackTrace();
            status.setException(e);
        } catch (Exception e) {
            // e.printStackTrace();
            status.setException(e);
        }
        return status;
    }

    /**
     * Load options from configuration file.
     */
    private static String[] loadConfig(String[] args, String configFileName) throws IOException {
        if (configFileName == null) {
            return args;
        }
        BufferedReader br = new BufferedReader(new FileReader(configFileName));
        String line;
        List<String> oldArgs = new ArrayList<String>(Arrays.asList(args));
        List<String> newArgs = new ArrayList<String>(Arrays.asList(args));
        while ((line = br.readLine()) != null) {
            String[] parts = line.split("=");
            if (parts.length > 2) {
                throw new RuntimeException();
            }
            String key = OptConfig.LONG_ARG_PREFIX + parts[0].trim();
            if (!oldArgs.contains(key)) {
                newArgs.add(key);
                if (parts.length > 1) {
                    newArgs.add(parts[1].trim());
                }
            }
        }
        br.close();
        return newArgs.toArray(new String[0]);
    }

    /**
     * Prints help message based on the given object.
     */
    public static OptParserStatus printHelp(Object options) {
        if (options == null) {
            throw new NullPointerException();
        }
        OptParserStatus status = new OptParserStatus();
        try {
            List<Field> argFields = findArgFields(options);
            checkArgsDef(options, argFields);
            assignAnnotations(options, argFields);
            StringBuilder helpMsg = new StringBuilder();
            if (options.getClass().isAnnotationPresent(OptHelp.class)) {
                helpMsg.append(options.getClass().getAnnotation(OptHelp.class).value());
            } else {
                helpMsg.append(options.getClass());
            }
            helpMsg.append("\n\n");
            helpMsg.append(HELP_ARG.getHelp()).append("\n\n");
            helpMsg.append("\n\n");
            helpMsg.append(CONFIG_ARG.getHelp()).append("\n\n");
            for (Field argField : argFields) {
                Opt<?> arg = (Opt<?>) argField.get(options);
                helpMsg.append(arg.getHelp());
                helpMsg.append("\n\n");
            }
            helpMsg.append(OptConfig.REQUIRED_HELP).append(' ').append("Required arguments.").append('\n');
            System.out.print(helpMsg);
        } catch (OptException e) {
            status.setException(e);
        } catch (Exception e) {
            status.setException(e);
        }
        return status;
    }

    /**
     * Checks if there is (default) help argument on the command line.
     */
    public static boolean hasHelpArgument(String[] args) {
        for (String s : args) {
            if (s.equals(HELP_ARG.getName()) || s.equals(HELP_ARG.getShortName())) {
                return true;
            }
        }
        return false;
    }

    /**
     * Configuration file name. If file is not specified returns null.
     */
    private static String configFileName(String[] args) {
        for (int i = 0; i < args.length; i++) {
            if (args[i].equals(CONFIG_ARG.getName()) || args[i].equals(CONFIG_ARG.getShortName())) {
                return args[i + 1];
            }
        }
        return null;
    }

    /**
     * Prints help message if help option is provided on the command line,
     * otherwise parses the command line options and saves the values in the
     * given object. The return value can be used to check if parsing was
     * successful.
     */
    public static OptParserStatus helpOrParse(Object options, String[] args) {
        OptParserStatus status;
        if (hasHelpArgument(args)) {
            status = printHelp(options);
        } else {
            status = parse(options, args);
        }
        return status;
    }

    /**
     * Finds the {@link Opt}s that are defined in the given object.
     */
    private static List<Field> findArgFields(Object obj) {
        List<Field> argFields = new LinkedList<Field>();
        Class<?> clz = obj.getClass();
        while (clz != null && !clz.equals(Object.class)) {
            Field[] fields = clz.getDeclaredFields();
            for (Field field : fields) {
                field.setAccessible(true);
                if (Opt.class.isAssignableFrom(field.getType())) {
                    argFields.add(field);
                }
            }
            clz = clz.getSuperclass();
        }
        return argFields;
    }

    /**
     * Checks correctness of the command line: (1) duplicate arguments, (2)
     * non-existing argument.
     */
    private static void checkCmd(Object obj, String[] cmd, List<Field> argFields) throws IllegalArgumentException,
            IllegalAccessException, OptException {
        // collect short and long names
        Set<String> names = new HashSet<String>();
        names.add(HELP_ARG.getName());
        names.add(HELP_ARG.getShortName());
        names.add(CONFIG_ARG.getName());
        names.add(CONFIG_ARG.getShortName());

        for (Field argField : argFields) {
            Opt<?> arg = (Opt<?>) argField.get(obj);
            names.add(arg.getName());
            if (arg.getShortName() != null) {
                names.add(arg.getShortName());
            }
        }
        // collect duplicate and collect name
        Set<String> cmdNames = new HashSet<String>();
        for (String s : cmd) {
            if (s.startsWith(OptConfig.LONG_ARG_PREFIX)
                    || (s.startsWith(OptConfig.SHORT_ARG_PREFIX) && isNotANumber(s))) {
                if (!cmdNames.add(s)) {
                    throw new OptException(s + " is duplicate on command line.");
                }
            }
        }
        // check non-existing argument
        for (String s : cmdNames) {
            if (!names.contains(s)) {
                throw new OptException(s + " is not a valid argument.");
            }
        }
    }

    private static boolean isNotANumber(String s) {
        try {
            Integer.parseInt(s);
        } catch (NumberFormatException ex1) {
            try {
                Double.parseDouble(s);
            } catch (NumberFormatException ex2) {
                try {
                    Long.parseLong(s);
                } catch (NumberFormatException ex3) {
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Checks if arguments are properly defined: (1) {@link Opt} fields are
     * final and non-null; and (2) no two arguments are with the same name.
     */
    private static void checkArgsDef(Object obj, List<Field> argFields) throws OptException, IllegalArgumentException,
            IllegalAccessException {
        // check fields are final
        for (Field field : argFields) {
            if (!Modifier.isFinal(field.getModifiers())) {
                throw new OptException(Opt.class.getName() + " fields must be final.");
            }
        }
        // check fields are non-null
        for (Field field : argFields) {
            if (field.get(obj) == null) {
                throw new OptException(Opt.class.getName() + " fields must not be null.");
            }
        }
        // check there are no two arguments with the same name
        Set<String> names = new HashSet<String>();
        Set<String> shortNames = new HashSet<String>();
        for (Field field : argFields) {
            Opt<?> arg = (Opt<?>) field.get(obj);
            // can check null
            if (!names.add(arg.getName())) {
                throw new OptException(arg.getName() + " is not unique.");
            }
            if (arg.getShortName() != null && !shortNames.add(arg.getShortName())) {
                throw new OptException(arg.getShortName() + " is not unique.");
            }
        }
    }

    /**
     * Assign fields of {@link Opt}s based on the present annotations.
     */
    private static void assignAnnotations(Object obj, List<Field> argFields) throws IllegalArgumentException,
            IllegalAccessException {
        for (Field argField : argFields) {
            Opt<?> arg = (Opt<?>) argField.get(obj);
            if (argField.isAnnotationPresent(OptHelp.class)) {
                arg.setHelpMsg(argField.getAnnotation(OptHelp.class).value());
            }
            if (argField.isAnnotationPresent(OptRequired.class)) {
                arg.setRequired(true);
            }
        }
    }

    /**
     * Assigns values to arguments based on the command line.
     */
    private static void assignValues(Object obj, String[] cmd, List<Field> argFields) throws IllegalArgumentException,
            IllegalAccessException, OptException, SecurityException, NoSuchFieldException {
        List<Field> assignFields = new ArrayList<Field>();
        for (Field argField : argFields) {
            Opt<?> arg = (Opt<?>) argField.get(obj);
            for (int i = 0; i < cmd.length; i++) {
                if (cmd[i].equals(arg.getName()) || cmd[i].equals(arg.getShortName())) {
                    arg.parseValue(cmd, i);
                    assignFields.add(argField);
                    // Field valueField = Arg.class.getDeclaredField("value");
                    // valueField.setAccessible(true);
                    // assignArgValue(arg, valueField, cmd, i);
                }
            }
        }
        // check if required values are set
        for (Field argField : argFields) {
            Opt<?> arg = (Opt<?>) argField.get(obj);
            if (arg.isRequired() && !assignFields.contains(argField)) {
                throw new OptException(arg.getName() + " is required.");
            }
        }
    }

    // /**
    // * Checks if values are properly assigned: mandatory arguments have
    // values.
    // */
    // private static void checkArgsValues(Object obj, List<Field> argFields)
    // throws IllegalArgumentException,
    // IllegalAccessException, OptException {
    // for (Field argField : argFields) {
    // Opt<?> arg = (Opt<?>) argField.get(obj);
    // if (arg.val() == null && arg.isRequired()) {
    // throw new OptException(arg.getName() + " is required.");
    // }
    // }
    // }

    // private static void assignArgValue(Arg<?> arg, Field valueField, String[]
    // cmd, int index)
    // throws IllegalAccessException, ArgException {
    // if (arg.getType() == boolean.class || arg.getType() == Boolean.class) {
    // valueField.set(arg, new Boolean(true));
    // } else if (arg.getType() == byte.class || arg.getType() == Byte.class) {
    // valueField.set(arg, new Byte(Byte.parseByte(cmd[index + 1])));
    // } else if (arg.getType() == short.class || arg.getType() == Short.class)
    // {
    // valueField.set(arg, new Short(Short.parseShort(cmd[index + 1])));
    // } else if (arg.getType() == int.class || arg.getType() == Integer.class)
    // {
    // valueField.set(arg, new Integer(Integer.parseInt(cmd[index + 1])));
    // } else if (arg.getType() == long.class || arg.getType() == Long.class) {
    // valueField.set(arg, new Long(Long.parseLong(cmd[index + 1])));
    // } else if (arg.getType() == float.class || arg.getType() == Float.class)
    // {
    // valueField.set(arg, new Float(Float.parseFloat(cmd[index + 1])));
    // } else if (arg.getType() == double.class || arg.getType() ==
    // Double.class) {
    // valueField.set(arg, new Double(Double.parseDouble(cmd[index + 1])));
    // } else if (arg.getType() == String.class) {
    // valueField.set(arg, cmd[index + 1]);
    // // } else if (arg.getType() == String[].class) {
    // // String[] values = cmd[index + 1].split(COLLECTION_SEPARATOR);
    // // valueField.set(arg, values);
    // } else if (arg.getType() == List.class) {
    // String[] values = cmd[index + 1].split(COLLECTION_SEPARATOR);
    // valueField.set(arg, Arrays.asList(values));
    // } else {
    // throw new ArgException("Type " + arg.getType().getName() +
    // " is not supported.");
    // }
    // }

    /**
     * Executes all methods in the given object that are annotated with
     * {@link OptPostprocess}.
     * 
     * @throws InvocationTargetException
     */
    private static void executeArgAfter(Object obj) throws IllegalArgumentException, IllegalAccessException,
            OptException, InvocationTargetException {
        List<Method> argAfters = findPostprocessors(obj);
        for (Method method : argAfters) {
            try {
                method.invoke(obj);
            } catch (InvocationTargetException e) {
                if (e.getCause() instanceof OptException) {
                    throw ((OptException) e.getCause());
                } else {
                    throw e;
                }
            }
        }
    }

    /**
     * Finds all methods in the given object that are annotated with
     * {@link OptPostprocess}.
     */
    private static List<Method> findPostprocessors(Object obj) {
        List<Method> argAfters = new LinkedList<Method>();
        Class<?> clz = obj.getClass();
        while (clz != null && !clz.equals(Object.class)) {
            Method[] methods = clz.getDeclaredMethods();
            for (Method method : methods) {
                method.setAccessible(true);
                if (method.isAnnotationPresent(OptPostprocess.class) && method.getTypeParameters().length == 0) {
                    argAfters.add(method);
                }
            }
            clz = clz.getSuperclass();
        }
        return argAfters;
    }

    public static void main(String... args) {
        @SuppressWarnings("unused")
        @OptHelp("describe class")
        class C {
            @OptHelp("describe boolean")
            final BooleanOpt zArg = new BooleanOpt("z");

            @OptRequired
            @OptHelp("describe integer")
            final IntegerOpt iArg = new IntegerOpt("integer", "i");
        }
        C obj = new C();
        printHelp(obj);
    }
}
