package coco.runtime;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Stack;
import java.util.TreeMap;

/**
 * Saves states that are encountered during the execution. This should be the
 * only class shared between user's application and our library.
 * 
 * This class is thread safe. Synchronized collections are used to keep track of
 * states and there is a thread local copy of an object to keep track of the
 * status of the execution (for example, thread is in predicate and similar).
 * 
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 * @author Rohan Sharma (sharma.rohan1990@gmail.com)
 */
public class CoCoRuntime {

    public enum Coverage {
        NONE, SC, BC, PATH, TIMP, PCTMS, PCTBB, PCTST
    }

    /** Systems tmp directory */
    private static final String TMP_DIR = "/tmp/";

    /** Mapping from test to states */
    private static final String TEST_TO_STATES_FILE = "test-to-states.txt";

    /** States that are touched */
    private static final String STATES_FILE = "touch-log.txt";

    /** File that indicates that path coverage should be used */
    private static final String PATH_FLAG_FILE = "path.coco.flag";
    /** File that indicates that only ids should be collected for touchBranch */
    private static final String BC_FLAG_FILE = "bc.coco.flag";
    /** File that indicates that only ids should be collected for each touch */
    private static final String SC_FLAG_FILE = "sc.coco.flag";
    /** File that indicates that PCT(BB) is executed */
    private static final String PCT_BB_FLAG_FILE = "pctbb.coco.flag";
    /** File that indicates that PCT(ST) is executed */
    private static final String PCT_ST_FLAG_FILE = "pctst.coco.flag";
    /** File that indicates that PCT(MS) is executed */
    private static final String PCT_MS_FLAG_FILE = "pctms.coco.flag";
    /** File that indicates that TIMP is executed */
    private static final String TIMP_FLAG_FILE = "timp.coco.flag";

    /** True if the execution should collect path coverage */
    private boolean isPath;
    /** Paths for methods that are currently on stack */
    private Stack<StringBuilder> paths = new Stack<StringBuilder>();

    /** True if the execution should collect TIIMP coverage */
    private boolean isTIMP;
    /** Paths for methods that are currently on stack */
    private Stack<ArrayList<Integer>> timps = new Stack<ArrayList<Integer>>();

    /** True if execution should collect only ids for touchBranch */
    private boolean isBC;

    /** True if execution is for statement coverage */
    private boolean isSC;

    /** True if PCT(BB) is executing */
    private boolean isPCTBB;

    /** True if PCT(ST) is executing */
    private boolean isPCTST;

    /** True if PCT(MS) is executing */
    private boolean isPCTMS;

    /**
     * Type to store thread local info that are needed to know the current
     * status of a thread in a predicate.
     */
    private static class ThreadPCTStatus {
        /** If there was recursive call we should ignore invocation of touch */
        private boolean ignoreNextTouch;

        /** Flag used to prevent recursive calls from predicates */
        private int inPredicate;
    }

    public static final ThreadLocal<ThreadPCTStatus> predicateStatus = new ThreadLocal<CoCoRuntime.ThreadPCTStatus>() {
        @Override
        protected ThreadPCTStatus initialValue() {
            return new ThreadPCTStatus();
        }
    };

    /** Unique instance of this class */
    private static CoCoRuntime inst;

    /** Set of states that are encountered */
    // IMPORTANT: if class is changed see setUp method and change!
    private IStateSet states = new StateSet();

    private boolean isSaveTouchedToFile = true;

    /** Test methods to states that are encountered */
    private final Map<String, IStateSet> testToStates;

    private CoCoRuntime(Coverage coverage) {
        this.testToStates = new TreeMap<String, IStateSet>();

        this.isBC = initCoverage(BC_FLAG_FILE) || (coverage == Coverage.BC);
        this.isPath = initCoverage(PATH_FLAG_FILE) || (coverage == Coverage.PATH);
        this.isTIMP = initCoverage(TIMP_FLAG_FILE) || (coverage == Coverage.TIMP);
        this.isPCTBB = initCoverage(PCT_BB_FLAG_FILE) || (coverage == Coverage.PCTBB);
        this.isPCTMS = initCoverage(PCT_MS_FLAG_FILE) || (coverage == Coverage.PCTMS);
        this.isPCTST = initCoverage(PCT_ST_FLAG_FILE) || (coverage == Coverage.PCTST);
        this.isSC = initCoverage(SC_FLAG_FILE) || (coverage == Coverage.SC);

        int numOfFlags = 0;
        numOfFlags += this.isBC ? 1 : 0;
        numOfFlags += this.isPath ? 1 : 0;
        numOfFlags += this.isTIMP ? 1 : 0;
        numOfFlags += this.isPCTBB ? 1 : 0;
        numOfFlags += this.isPCTMS ? 1 : 0;
        numOfFlags += this.isPCTST ? 1 : 0;
        numOfFlags += this.isSC ? 1 : 0;

        if (numOfFlags != 1) {
            throw new RuntimeException("Coverage flag is not set!");
        }

        printExecutionMode();

        Runtime.getRuntime().addShutdownHook(new Thread() {
            @Override
            public void run() {
                IStateSet allStates = states;
                // need to collect all states if we have separate set for
                // multiple test cases
                if (!testToStates.isEmpty() && states instanceof StateSet) {
                    allStates = new StateSet(); // TODO: introduce a factory
                    for (Entry<String, IStateSet> en : testToStates.entrySet()) {
                        for (String state : en.getValue().states()) {
                            allStates.add(state);
                        }
                    }
                }
                if (isSaveTouchedToFile) {
                    System.out.println("NumOfStates " + allStates.size());
                }
                if (isSaveTouchedToFile && allStates.size() > 0) {
                    try {
                        PrintWriter pw = new PrintWriter(new FileWriter(STATES_FILE, true), true);
                        for (String s : allStates.states()) {
                            pw.println(s);
                            // pw.println(s.hashCode());
                        }
                        pw.close();
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                }
                if (!testToStates.isEmpty()) {
                    try {
                        PrintWriter pw = new PrintWriter(new FileWriter(TEST_TO_STATES_FILE, true), true);
                        for (Entry<String, IStateSet> en : testToStates.entrySet()) {
                            // connect all states
                            StringBuilder sb = new StringBuilder();
                            for (String state : en.getValue().states()) {
                                if (sb.length() == 0) {
                                    sb.append(state);
                                } else {
                                    sb.append(",").append(state);
                                }
                            }
                            pw.println(en.getKey() + " >>" + sb + "<<");
                        }
                        if (pw != null)
                            pw.close();
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                }
            }
        });
    }

    public synchronized static CoCoRuntime inst() {
        return inst(Coverage.NONE);
    }

    public synchronized static CoCoRuntime inst(Coverage coverage) {
        if (inst == null) {
            inst = new CoCoRuntime(coverage);
        }
        return inst;
    }

    /**
     * This method should be invoked before each test case. Note that we do not
     * rely on this method because invocation is not required. However the plan
     * is to use it for non-data-structures.
     */
    public final void setUp(String testName) {
        if (testToStates.containsKey(testName)) {
            states = testToStates.get(testName);
        } else {
            states = new StateSet();
            testToStates.put(testName, states);
        }
    }

    public final void setSaveTouchedToFile(boolean f) {
        isSaveTouchedToFile = f;
    }

    private final void touchMe(String prefix, int pointId, int... predicates) {
        if (predicateStatus.get().ignoreNextTouch) {
            predicateStatus.get().ignoreNextTouch = false;
            return;
        }

        Arrays.sort(predicates);

        StringBuilder sb = new StringBuilder(prefix);
        sb.append(pointId);
        for (int predicateId : predicates) {
            // TODO: try binary search (or similar) if necessary
            if (predicateId > 0) {
                sb.append('*').append(predicateId);
            }
        }

        // + new String(currentState));
        states.add(sb.toString());
    }

    public final void touch(int pointId, int... predicates) {
        if (isPath) {
            paths.peek().append(pointId).append('*');
        } else if (isTIMP) {
            checkAndAddTIMPState(pointId);
        } else if (isBC) {
            // NOP
        } else if (isSC) {
            touchMe("", pointId);
        } else {
            touchMe("", pointId, predicates);
        }
    }

    /**
     * This method is invoked for each branch.
     */
    public final void touchBranch(int pointId, int... predicates) {
        if (isPath) {
            paths.peek().append(pointId).append('*');
        } else if (isTIMP) {
            checkAndAddTIMPState(pointId);
        } else if (isBC) {
            touchMe("", pointId);
        } else if (isSC) {
            touchMe("", pointId);
        } else {
            touchMe("", pointId, predicates);
        }
    }

    /**
     * Invoked from instrumented code at the beginning of each method (/a basic
     * block).
     */
    public final void touchStartMethod(int pointId, int... predicates) {
        if (isPath) {
            paths.peek().append(pointId).append('*');
        } else if (isTIMP) {
            checkAndAddTIMPState(pointId);
        } else if (isBC) {
            touchMe("", pointId);
        } else if (isSC) {
            touchMe("", pointId);
        } else {
            touchMe("", pointId, predicates);
        }
    }

    /**
     * Marker for the beginning of a static block.
     */
    public final void touchBeginStatic() {
        touchBeginMethod();
    }

    /**
     * Marker for the end of a static block.
     */
    public final void touchEndStatic() {
        touchEndMethod();
    }

    /**
     * Marker for the beginning of a method.
     */
    public final void touchBeginMethod() {
        if (isPath) {
            paths.push(new StringBuilder());
        } else if (isTIMP) {
            timps.push(new ArrayList<Integer>());
        }
    }

    /**
     * Marker for the end of a method.
     */
    public final void touchEndMethod() {
        if (isPath) {
            StringBuilder state = paths.pop();
            if (state.length() > 0) {
                states.add(state.toString());
            }
        } else if (isTIMP) {
            addTIMPState(timps.pop());
        }
    }

    /**
     * Invoked from instrumented code to avoid recursive calls to predicates.
     */
    public final boolean testAndSetInPredicate() {
        predicateStatus.get().inPredicate++;
        if (predicateStatus.get().inPredicate == 1) {
            return false;
        }
        return true;
    }

    /**
     * Invoked from instrumented code to avoid recursive calls to predicates.
     */
    public final void resetInPredicate() {
        predicateStatus.get().inPredicate--;
    }

    /**
     * Invoked from instrumented code to ignore next touch to avoid infinite
     * loops.
     */
    public final void ignoreNextTouch() {
        predicateStatus.get().ignoreNextTouch = true;
    }

    public final void setPathCoverage(boolean f) {
        isPath = f;
    }

    /**
     * Return all currently collected states.
     */
    final IStateSet getStates() {
        return states;
    }

    // PRIVATE

    /**
     * Print execution mode.
     */
    private void printExecutionMode() {
        if (isBC)
            System.out.println("Running BC");
        if (isPath)
            System.out.println("Running PATH");
        if (isTIMP)
            System.out.println("Running TIMP");
        if (isPCTBB)
            System.out.println("Running PCT(BB)");
        if (isPCTMS)
            System.out.println("Running PCT(MS)");
        if (isPCTST)
            System.out.println("Running PCT(ST)");
        if (isSC)
            System.out.println("Running SC");
    }

    private boolean initCoverage(String fileName) {
        return new File(fileName).exists() || new File(TMP_DIR + fileName).exists();
    }

    /**
     * Check if point is already in the state and if it save current state and
     * start new state; if it is not, just add the point to the current state.
     */
    private void checkAndAddTIMPState(int pointId) {
        if (timps.peek().contains(pointId)) {
            addTIMPState(timps.pop());
            timps.add(new ArrayList<Integer>());
        }
        timps.peek().add(pointId);
    }

    private void addTIMPState(ArrayList<Integer> vals) {
        if (vals.size() > 0) {
            StringBuilder sb = new StringBuilder();
            for (Integer el : vals) {
                sb.append(el).append('*');
            }
            if (sb.length() > 0) {
                states.add(sb.toString());
            }
        }
    }
}
