package coco.runtime;

import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Iterator;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import newcommand.Newcommand;

import org.junit.runner.Description;
import org.junit.runner.notification.Failure;
import org.junit.runner.notification.RunListener;

/**
 * @author Milos Gligoric (milos.gligoric@gmail.com)
 * @author Rohan Sharma (sharma.rohan1990@gmail.com)
 */
class CoCoRuntimeListener extends RunListener {

    /** Sign between key and value */
    private static final String SIGN_BTW = "=";

    /** Separator of states */
    private static final String COMMA_LITERAL = ",";

    /** Marker before the first state */
    private static final String MARKER_BEFORE = ">>";

    /** Marker after the last state */
    private static final String MARKER_AFTER = "<<";

    /**
     * Sorting test names.
     */
    public static class CompareTestNames implements Comparator<String> {
        private static final String nameString = ".*_(.*)_.*test(.*) .*";
        private static final Pattern namePattern = Pattern.compile(nameString);

        @Override
        public int compare(String testName0, String testName1) {
            Matcher m0 = namePattern.matcher(testName0);
            Matcher m1 = namePattern.matcher(testName1);
            m0.find();
            m1.find();
            int fileId0 = Integer.parseInt(m0.group(1));
            int testId0 = Integer.parseInt(m0.group(2));
            int fileId1 = Integer.parseInt(m1.group(1));
            int testId1 = Integer.parseInt(m1.group(2));
            if (fileId0 > fileId1) {
                return 1;
            } else if (fileId0 < fileId1) {
                return -1;
            } else {
                return testId0 - testId1;
            }
        }
    }

    private TreeSet<String> testInfos = new TreeSet<String>(new CompareTestNames());

    private CoCoRuntimeOpt options;

    private long startTime;

    private int testClassCounter;
    private int testCaseCounter;
    private int failureCaseCounter;

    private boolean isSaveCoveragePerTest;
    private PrintWriter saveCoveragePerTestPW;

    private TreeSet<String> allStates = new TreeSet<String>();

    CoCoRuntimeListener(CoCoRuntimeOpt options) {
        this.isSaveCoveragePerTest = options.getSaveCoveragePerTestFileName() != null;
        if (isSaveCoveragePerTest) {
            try {
                saveCoveragePerTestPW = new PrintWriter(options.getSaveCoveragePerTestFileName());
            } catch (FileNotFoundException e) {
                e.printStackTrace();
            }
        }
        this.options = options;
    }

    public void bcRunStarted() {
        startTime = System.currentTimeMillis();
    }

    public void bcRunFinished() {
        long executionTime = System.currentTimeMillis() - startTime;
        if (options.isUseNewcommand()) {
            println(Newcommand.makeNewcommand(prefixMacroSuffix(options.getStatesText()), allStates.size()));
        } else {
            String s = String.format("%s %s %d", options.getStatesText(), SIGN_BTW, allStates.size());
            println(prefixMacroSuffix(s));
        }
        if (options.isUseNewcommand()) {
            println(Newcommand.makeNewcommand(prefixMacroSuffix(options.getNewcommandTimeText()), executionTime));
        } else {
            String s = String.format("%s %s %d", options.getNewcommandTimeText(), SIGN_BTW, executionTime);
            println(prefixMacroSuffix(s));
        }
        if (options.isUseNewcommand()) {
            println(Newcommand
                    .makeNewcommand(prefixMacroSuffix(options.getNewcommandTestClassText()), testClassCounter));
        } else {
            String s = String.format("%s %s %d", options.getNewcommandTestClassText(), SIGN_BTW, testClassCounter);
            println(prefixMacroSuffix(s));
        }
        if (options.isUseNewcommand()) {
            println(Newcommand.makeNewcommand(prefixMacroSuffix(options.getNewcommandTestCaseText()), testCaseCounter));
        } else {
            String s = String.format("%s %s %d", options.getNewcommandTestCaseText(), SIGN_BTW, testCaseCounter);
            println(prefixMacroSuffix(s));
        }
        if (options.isUseNewcommand()) {
            println(Newcommand.makeNewcommand(prefixMacroSuffix(options.getNewcommandFailureCaseText()),
                    failureCaseCounter));
        } else {
            String s = String.format("%s %s %d", options.getNewcommandFailureCaseText(), SIGN_BTW, failureCaseCounter);
            println(prefixMacroSuffix(s));
        }

        int counter = 0;
        if (isSaveCoveragePerTest) {
            Iterator<String> iterator = testInfos.iterator();
            while (iterator.hasNext()) {
                saveCoveragePerTestPW.println(counter + " " + iterator.next());
                counter++;
            }
            saveCoveragePerTestPW.flush();
            saveCoveragePerTestPW.close();
        }
        testInfos.clear();
    }

    private String prefixMacroSuffix(String macro) {
        return options.getNewcommandPrefix() + macro + options.getNewcommandSuffix();
    }

    private void println(String s) {
        System.out.println(s);
    }

    /* RunListener interface */

    @Override
    public void testRunStarted(Description descripton) {
        testClassCounter++;
    }

    @Override
    public void testFailure(Failure failure) {
        failureCaseCounter++;
    }

    @Override
    public void testStarted(Description description) {
        testCaseCounter++;
    }

    @Override
    public void testFinished(Description description) {
        allStates.addAll(Arrays.asList(CoCoRuntime.inst().getStates().states()));
        if (isSaveCoveragePerTest) {
            StringBuilder sb = new StringBuilder();
            sb.append(description.getClassName() + "." + description.getMethodName() + " ");
            String statesAsString = "";
            String[] statesAsStringArray = CoCoRuntime.inst().getStates().states();
            for (int i = 0; i < statesAsStringArray.length; i++) {
                statesAsString += statesAsStringArray[i];
                if (i != statesAsStringArray.length - 1) {
                    statesAsString += COMMA_LITERAL;
                }
            }
            sb.append(MARKER_BEFORE + statesAsString + MARKER_AFTER);
            testInfos.add(sb.toString());

            // clear states for the next test
            CoCoRuntime.inst().getStates().clear();
        }
    }

}
