package example.avl;

public class IntAVLTreeMap {
    private Node root;
    private int size;

    public class Node {
        int key;
        Object value;
        Node left;
        Node right;
        int height;

        @Override
        public String toString() {
            String s = "{c:" + ";v" + value + "[" + (left != null ? left.toString() : null) + "]["
                    + (right != null ? right.toString() : null) + "]}";
            return s;
        }
    }

    @Override
    public String toString() {
        if (root == null) {
            return "null";
        } else {
            return root.toString();
        }
    }

    public boolean containsKey(Integer key) {
        return findNode(key) != null;
    }

    public Object get(Integer key) {
        Node x = findNode(key);
        return x == null ? null : x.value;
    }

    public void put(Integer key, Object value) {
        Node x = new Node();
        x.key = key;
        x.value = value;
        treeInsert(x);
    }

    public Object remove(Integer key) {
        Node fnd = findNode(key);
        if (fnd == null) {
            return null;
        }
        Node ret = treeDelete(fnd);
        size--;
        return ret.value;
    }

    public int getSize() {
        return size;
    }

    private int getHeight(Node x) {
        return x == null ? -1 : x.height;
    }

    private int getBalance(Node x) {
        return x == null ? 0 : getHeight(x.left) - getHeight(x.right);
    }

    private Node findNode(Integer key) {
        Node cur = root;
        while (cur != null && key.compareTo(cur.key) != 0) {
            if (key.compareTo(cur.key) < 0) {
                cur = cur.left;
            } else {
                cur = cur.right;
            }
        }
        return cur;
    }

    private Node rightRotate(Node x) {
        if (x == null) {
            return null;
        }
        Node y = x.left;
        x.left = y.right;
        y.right = x;
        x.height = 1 + Math.max(getHeight(x.left), getHeight(x.right));
        y.height = 1 + Math.max(getHeight(y.left), getHeight(x));
        return y;
    }

    private Node leftRotate(Node x) {
        if (x == null) {
            return null;
        }
        Node y = x.right;
        x.right = y.left;
        y.left = x;
        x.height = 1 + Math.max(getHeight(x.left), getHeight(x.right));
        y.height = 1 + Math.max(getHeight(y.right), getHeight(x));
        return y;
    }

    private Node rightLeftRotate(Node x) {
        x.right = rightRotate(x.right);
        return leftRotate(x);
    }

    private Node leftRightRotate(Node x) {
        x.left = leftRotate(x.left);
        return rightRotate(x);
    }

    public void treeInsert(Node x) {
        Node fnd = findNode(x.key);
        if (fnd != null) {
            fnd.key = x.key;
            return;
        }
        root = treeInsert_recur(x, root);
        size++;
    }

    private Node treeInsert_recur(Node x, Node cur) {
        if (cur == null) {
            return x;
        }
        if (((Integer) x.key).compareTo(cur.key) < 0) {
            cur.left = treeInsert_recur(x, cur.left);
        } else {
            if (((Integer) x.key).compareTo(cur.key) == 0) {
                cur.value = x.value;
                return cur;
            } else {
                cur.right = treeInsert_recur(x, cur.right);
            }
        }

        cur.height = 1 + Math.max(getHeight(cur.left), getHeight(cur.right));
        int balance = getBalance(cur);
        if (balance == -2) {
            if (getBalance(cur.right) == -1) {
                return leftRotate(cur);
            } else {
                if (getBalance(cur.right) == 1) {
                    return rightLeftRotate(cur);
                }
            }
        } else {
            if (balance == 2) {
                if (getBalance(cur.left) == 1) {
                    return rightRotate(cur);
                } else {
                    if (getBalance(cur.left) == -1) {
                        return leftRightRotate(cur);
                    }
                }
            }
        }
        return cur;
    }

    private Node treeDelete(Node x) {
        root = treeDelete_recur(x, root);
        return x;
    }

    private Node treeDelete_recur(Node x, Node cur) {
        if (((Integer) x.key).compareTo(cur.key) < 0) {
            cur.left = treeDelete_recur(x, cur.left);
        } else {
            if (((Integer) x.key).compareTo(cur.key) == 0) {
                return afterDelete(cur);
            } else {
                cur.right = treeDelete_recur(x, cur.right);
            }
        }
        return deleteFix(cur);
    }

    private Node deleteFix(Node cur) {
        cur.height = 1 + Math.max(getHeight(cur.left), getHeight(cur.right));
        int balance = getBalance(cur);
        if (balance == -2) {
            if (getBalance(cur.right) <= 0) {
                return leftRotate(cur);
            } else {
                if (getBalance(cur.right) == 1) {
                    return rightLeftRotate(cur);
                }
            }
        } else {
            if (balance == 2) {
                if (getBalance(cur.left) >= 0) {
                    return rightRotate(cur);
                } else {
                    if (getBalance(cur.left) == -1) {
                        return leftRightRotate(cur);
                    }
                }
            }
        }
        return cur;
    }

    private Node afterDelete(Node x) {
        if (x.left == null && x.right == null) {
            return null;
        } else {
            if (x.left != null && x.right != null) {
                Node z = getIOS(x);
                x.right = treeDelete_recur(z, x.right);
                x.key = z.key;
                return deleteFix(x);
            } else {
                if (x.left != null) {
                    return x.left;
                } else {
                    return x.right;
                }
            }
        }
    }

    private Node getIOS(Node z) {
        Node x, y = null;
        x = z.right;
        while (x != null) {
            y = x;
            x = x.left;
        }
        return y;
    }

    protected int compare(Integer k1, Integer k2) {
        return k1.compareTo(k2);
    }

}
