package example.fibonacciheap;

import java.util.ArrayList;
import java.util.List;
import java.util.Stack;

public class FibonacciHeap {

    private final static double oneOverLogPhi = 1.0 / Math.log((1.0 + Math.sqrt(5.0)) / 2.0);

    private final static int FALSE = 0;
    private final static int TRUE = 1;

    public FibonacciHeapNode minNode;

    public int size;

    public void delete(FibonacciHeapNode x) {
        // make x as small as possible
        decreaseKey(x, Integer.MIN_VALUE);
        // remove the smallest, which decreases n also
        removeMin();
    }

    public void decreaseKey(FibonacciHeapNode x, int k) {
        if (k > x.key) {
            throw new IllegalArgumentException("decreaseKey() got larger key value");
        }
        x.key = k;
        FibonacciHeapNode y = x.parent;
        if ((y != null) && (x.key < y.key)) {
            cut(x, y);
            cascadingCut(y);
        }
        if (x.key < minNode.key) {
            minNode = x;
        }
    }

    protected void cascadingCut(FibonacciHeapNode y) {
        FibonacciHeapNode z = y.parent;
        if (z != null) {
            if (y.mark == FALSE) {
                y.mark = TRUE;
            } else {
                cut(y, z);
                cascadingCut(z);
            }
        }
    }

    protected void cut(FibonacciHeapNode x, FibonacciHeapNode y) {
        x.left.right = x.right;
        x.right.left = x.left;
        y.degree--;
        if (y.child == x) {
            y.child = x.right;
        }
        if (y.degree == 0) {
            y.child = null;
        }
        x.left = minNode;
        x.right = minNode.right;
        minNode.right = x;
        x.right.left = x;
        x.parent = null;
        x.mark = FALSE;
    }

    public FibonacciHeapNode removeMin() {
        FibonacciHeapNode z = minNode;
        if (z != null) {
            int numKids = z.degree;
            FibonacciHeapNode x = z.child;
            FibonacciHeapNode tempRight;
            while (numKids > 0) {
                tempRight = x.right;
                x.left.right = x.right;
                x.right.left = x.left;
                x.left = minNode;
                x.right = minNode.right;
                minNode.right = x;
                x.right.left = x;
                x.parent = null;
                x = tempRight;
                numKids--;
            }
            z.left.right = z.right;
            z.right.left = z.left;
            if (z == z.right) {
                minNode = null;
            } else {
                minNode = z.right;
                consolidate();
            }
            size--;
        }
        return z;
    }

    protected void consolidate() {
        int arraySize = ((int) Math.floor(Math.log(size) * oneOverLogPhi)) + 1;
        // <FibonacciHeapNode>
        List<FibonacciHeapNode> array = new ArrayList<FibonacciHeapNode>(arraySize);
        for (int i = 0; i < arraySize; i++) {
            array.add(null);
        }
        int numRoots = 0;
        FibonacciHeapNode x = minNode;
        if (x != null) {
            numRoots++;
            x = x.right;
            while (x != minNode) {
                numRoots++;
                x = x.right;
            }
        }

        while (numRoots > 0) {
            int d = x.degree;
            FibonacciHeapNode next = x.right;
            for (;;) {
                FibonacciHeapNode y = (FibonacciHeapNode) array.get(d);
                if (y == null) {
                    break;
                }
                if (x.key > y.key) {
                    FibonacciHeapNode temp = y;
                    y = x;
                    x = temp;
                }
                link(y, x);
                array.set(d, null);
                d++;
            }

            array.set(d, x);
            x = next;
            numRoots--;
        }

        minNode = null;
        for (int i = 0; i < arraySize; i++) {
            FibonacciHeapNode y = (FibonacciHeapNode) array.get(i);
            if (y == null) {
                continue;
            }
            if (minNode != null) {
                y.left.right = y.right;
                y.right.left = y.left;
                y.left = minNode;
                y.right = minNode.right;
                minNode.right = y;
                y.right.left = y;
                if (y.key < minNode.key) {
                    minNode = y;
                }
            } else {
                minNode = y;
            }
        }
    }

    protected void link(FibonacciHeapNode y, FibonacciHeapNode x) {
        y.left.right = y.right;
        y.right.left = y.left;
        y.parent = x;
        if (x.child == null) {
            x.child = y;
            y.right = y;
            y.left = y;
        } else {
            y.left = x.child;
            y.right = x.child.right;
            x.child.right = y;
            y.right.left = y;
        }

        x.degree++;
        y.mark = FALSE;
    }

    public void insert(FibonacciHeapNode node, int key) {
        node.key = key;
        if (minNode != null) {
            node.left = minNode;
            node.right = minNode.right;
            minNode.right = node;
            node.right.left = node;
            if (key < minNode.key) {
                minNode = node;
            }
        } else {
            minNode = node;
        }
        size++;
    }

    @SuppressWarnings("unused")
    private int getMin() {
        FibonacciHeapNode temp = minNode;
        int min = minNode.key;
        do {
            if (temp.key < min) {
                min = temp.key;
            }
            temp = temp.right;
        } while (temp != minNode);
        return min;
    }

    @Override
    public String toString() {
        if (minNode == null) {
            return "FibonacciHeap=[]";
        }
        Stack<FibonacciHeapNode> stack = new Stack<FibonacciHeapNode>();
        stack.push(minNode);
        StringBuffer buf = new StringBuffer(512);
        buf.append("FibonacciHeap=[");
        while (!stack.empty()) {
            FibonacciHeapNode curr = (FibonacciHeapNode) stack.pop();
            buf.append(curr + (curr == minNode ? "MIN" : ""));
            buf.append(", ");
            if (curr.child != null) {
                stack.push(curr.child);
            }
            FibonacciHeapNode start = curr;
            curr = curr.right;
            while (curr != start) {
                buf.append(curr + (curr == minNode ? "MIN" : ""));
                buf.append(", ");
                if (curr.child != null) {
                    stack.push(curr.child);
                }
                curr = curr.right;
            }
        }

        buf.append(']');
        return buf.toString();
    }

    public boolean checkHeapified() {
        FibonacciHeapNode current = minNode;
        do {
            if (!current.checkHeapified()) {
                return false;
            }
            current = current.right;
        } while (current != minNode);
        return true;
    }

}
