package example.taco.cachelist;

public class NodeCachingLinkedList {

    public transient LinkedListNode header;
    protected transient int size;
    protected transient int modCount;

    protected final int DEFAULT_MAXIMUM_CACHE_SIZE = 20;

    protected transient LinkedListNode firstCachedNode;

    protected transient int cacheSize;

    protected int maximumCacheSize;

    protected LinkedListNode createHeaderNode() {
        return new LinkedListNode();
    }

    public NodeCachingLinkedList() {
        header = createHeaderNode();
        size = 0;
        cacheSize = 0;
        firstCachedNode = null;
        maximumCacheSize = DEFAULT_MAXIMUM_CACHE_SIZE;
    }

    @Override
    public String toString() {
        if (header == null) {
            return "null";
        } else {
            LinkedListNode c = header;
            String s = c.toString();
            while (c.next != header) {
                c = c.next;
                s += ":" + c.toString();
            }
            return s;
        }
    }

    protected int getMaximumCacheSize() {
        return maximumCacheSize;
    }

    protected void setMaximumCacheSize(int maximumCacheSize) {
        this.maximumCacheSize = maximumCacheSize;
        shrinkCacheToMaximumSize();
    }

    protected void shrinkCacheToMaximumSize() {
        while (cacheSize > maximumCacheSize) {
            getNodeFromCache();
        }
    }

    protected LinkedListNode getNodeFromCache() {
        if (cacheSize == 0) {
            return null;
        }

        LinkedListNode cachedNode = firstCachedNode;
        firstCachedNode = cachedNode.next;
        cachedNode.next = null;
        cacheSize--;
        return cachedNode;
    }

    protected boolean isCacheFull() {
        return cacheSize >= maximumCacheSize;
    }

    protected void addNodeToCache(LinkedListNode node) {
        if (isCacheFull()) {
            return;
        }

        LinkedListNode nextCachedNode = firstCachedNode;
        node.previous = null;
        node.next = nextCachedNode;
        node.setValue(null);
        firstCachedNode = node;
        cacheSize++;
    }

    protected LinkedListNode createNode(Object value) {
        LinkedListNode cachedNode = getNodeFromCache();
        if (cachedNode == null) {
            return super_createNode(value);
        } else {
            cachedNode.setValue(value);
            return cachedNode;
        }
    }

    private void super_removeNode(LinkedListNode node) {
        node.previous.next = node.next;
        node.next.previous = node.previous;
        size--;
        modCount++;
    }

    protected void removeNode(LinkedListNode node) {
        super_removeNode(node);
        addNodeToCache(node);
    }

    protected void removeAllNodes() {
        int numberOfNodesToCache = Math_min(size, maximumCacheSize - cacheSize);
        LinkedListNode node = header.next;
        for (int currentIndex = 0; currentIndex < numberOfNodesToCache; currentIndex++) {
            LinkedListNode oldNode = node;
            node = node.next;
            addNodeToCache(oldNode);
        }
        super_removeAllNodes();
    }

    private int Math_min(int left, int right) {
        return left < right ? left : right;
    }

    private LinkedListNode super_createNode(Object value) {
        return new LinkedListNode(value);
    }

    private void super_removeAllNodes() {
        header.next = header;
        header.previous = header;
        size = 0;
        modCount++;
    }

    public boolean remove(Object value) {
        for (LinkedListNode node = header.next; node != header; node = node.next) {
            if (isEqualValue(node.getValue(), value)) {
                removeNode(node);
                return true;
            }
        }
        return false;
    }

    protected boolean isEqualValue(Object value1, Object value2) {
        return (value1 == value2 || (value1 == null ? false : value1.equals(value2)));
    }

    public boolean add(Object value) {
        addLast(value);
        return true;
    }

    public boolean addLast(Object o) {
        addNodeBefore(header, o);
        return true;
    }

    protected void addNodeBefore(LinkedListNode node, Object value) {
        LinkedListNode newNode = createNode(value);
        addNode(newNode, node);
    }

    protected void addNode(LinkedListNode nodeToInsert, LinkedListNode insertBeforeNode) {
        nodeToInsert.next = insertBeforeNode;
        nodeToInsert.previous = insertBeforeNode.previous;
        insertBeforeNode.previous.next = nodeToInsert;
        insertBeforeNode.previous = nodeToInsert;
        size++;
        modCount++;
    }

    public boolean contains(Object arg) {
        return indexOf(arg) != -1;
    }

    public int indexOf(Object value) {
        int i = 0;
        for (LinkedListNode node = header.next; node != header; node = node.next) {
            if (isEqualValue(node.getValue(), value)) {
                return i;
            }
            i++;
        }
        return -1;
    }

    public Object removeIndex(int index) {
        LinkedListNode node = getNode(index, false);
        Object oldValue = node.getValue();
        removeNode(node);
        return oldValue;
    }

    protected LinkedListNode getNode(int index, boolean endMarkerAllowed) throws RuntimeException {
        if (index < 0) {
            throw new RuntimeException("Couldn't get the node: " + "index (" + index + ") less than zero.");
        }
        if (!endMarkerAllowed && index == size) {
            throw new RuntimeException("Couldn't get the node: " + "index (" + index + ") is the size of the list.");
        }
        if (index > size) {
            throw new IndexOutOfBoundsException("Couldn't get the node: " + "index (" + index
                    + ") greater than the size of the " + "list (" + size + ").");
        }

        LinkedListNode node;
        if (index < (size / 2)) {
            node = header.next;
            for (int currentIndex = 0; currentIndex < index; currentIndex++) {
                node = node.next;
            }
        } else {
            node = header;
            for (int currentIndex = size; currentIndex > index; currentIndex--) {
                node = node.previous;
            }
        }
        return node;
    }

}
