package example.taco.treeset;

public class TreeSet {

    private transient TreeSetEntry root = null;

    private transient int size = 0;

    private transient int modCount = 0;

    private final boolean RED = false;

    final boolean BLACK = true;

    @Override
    public String toString() {
        return root != null ? root.toString() : "null";
    }

    public boolean contains(int aKey) {
        return getEntry(aKey) != null;
    }

    private TreeSetEntry getEntry(int key) {
        TreeSetEntry p = root;
        while (p != null) {
            if (key == p.key) {
                return p;
            } else {
                if (key < p.key) {
                    p = p.left;
                } else {
                    p = p.right;
                }
            }
        }
        return null;
    }

    public boolean add(int aKey) {
        TreeSetEntry t = root;
        if (t == null) {
            incrementSize();
            root = new TreeSetEntry(aKey, null);
            return false;
        }

        while (true) {
            if (aKey == t.key) {
                return true;
            } else {
                if (aKey < t.key) {
                    if (t.left != null) {
                        t = t.left;
                    } else {
                        incrementSize();
                        t.left = new TreeSetEntry(aKey, t);
                        fixAfterInsertion(t.left);
                        return false;
                    }
                } else {
                    if (t.right != null) {
                        t = t.right;
                    } else {
                        incrementSize();
                        t.right = new TreeSetEntry(aKey, t);
                        fixAfterInsertion(t.right);
                        return false;
                    }
                }
            }
        }
    }

    private void incrementSize() {
        modCount++;
        size++;
    }

    private static boolean colorOf(TreeSetEntry p) {
        boolean black = true;
        return (p == null ? black : p.color);
    }

    private static TreeSetEntry parentOf(TreeSetEntry p) {
        return (p == null ? null : p.parent);
    }

    private void setColor(TreeSetEntry p, boolean c) {
        if (p != null) {
            p.color = c;
        }
    }

    private static TreeSetEntry leftOf(TreeSetEntry p) {
        return (p == null) ? null : p.left;
    }

    private static TreeSetEntry rightOf(TreeSetEntry p) {
        return (p == null) ? null : p.right;
    }

    private void rotateLeft(TreeSetEntry p) {
        TreeSetEntry r = p.right;
        p.right = r.left;
        if (r.left != null) {
            r.left.parent = p;
        }

        r.parent = p.parent;
        if (p.parent == null) {
            root = r;
        } else {
            if (p.parent.left == p) {
                p.parent.left = r;
            } else {
                p.parent.right = r;
            }
        }

        r.left = p;
        p.parent = r;
    }

    private void rotateRight(TreeSetEntry p) {
        TreeSetEntry l = p.left;
        p.left = l.right;
        if (l.right != null) {
            l.right.parent = p;
        }

        l.parent = p.parent;
        if (p.parent == null) {
            root = l;
        } else {
            if (p.parent.right == p) {
                p.parent.right = l;
            } else {
                p.parent.left = l;
            }
        }

        l.right = p;
        p.parent = l;
    }

    private void fixAfterInsertion(final TreeSetEntry entry) {
        TreeSetEntry x = entry;
        x.color = RED;
        while (x != null && x != root && x.parent.color == RED) {
            if (parentOf(x) == leftOf(parentOf(parentOf(x)))) {
                TreeSetEntry y = rightOf(parentOf(parentOf(x)));
                if (colorOf(y) == RED) {
                    setColor(parentOf(x), BLACK);
                    setColor(y, BLACK);
                    setColor(parentOf(parentOf(x)), RED);
                    x = parentOf(parentOf(x));
                } else {
                    if (x == rightOf(parentOf(x))) {
                        x = parentOf(x);
                        rotateLeft(x);
                    }
                    setColor(parentOf(x), BLACK);
                    setColor(parentOf(parentOf(x)), RED);
                    if (parentOf(parentOf(x)) != null) {
                        rotateRight(parentOf(parentOf(x)));
                    }
                }
            } else {
                TreeSetEntry y = leftOf(parentOf(parentOf(x)));
                if (colorOf(y) == RED) {
                    setColor(parentOf(x), BLACK);
                    setColor(y, BLACK);
                    setColor(parentOf(parentOf(x)), RED);
                    x = parentOf(parentOf(x));
                } else {
                    if (x == leftOf(parentOf(x))) {
                        x = parentOf(x);
                        rotateRight(x);
                    }
                    setColor(parentOf(x), BLACK);
                    setColor(parentOf(parentOf(x)), RED);
                    if (parentOf(parentOf(x)) != null) {
                        rotateLeft(parentOf(parentOf(x)));
                    }
                }
            }
        }

        root.color = BLACK;
    }

    public boolean remove(int aKey) {
        TreeSetEntry p = getEntry(aKey);
        if (p == null) {
            return false;
        }
        deleteEntry(p);
        return true;
    }

    private void deleteEntry(TreeSetEntry p) {
        decrementSize();
        if (p.left != null && p.right != null) {
            TreeSetEntry s = successor(p);
            p.key = s.key;
            p = s;
        }

        TreeSetEntry replacement = (p.left != null ? p.left : p.right);
        if (replacement != null) {
            replacement.parent = p.parent;
            if (p.parent == null) {
                root = replacement;
            } else {
                if (p == p.parent.left) {
                    p.parent.left = replacement;
                } else {
                    p.parent.right = replacement;
                }
            }

            p.left = p.right = p.parent = null;
            if (p.color == BLACK) {
                fixAfterDeletion(replacement);
            }
        } else {
            if (p.parent == null) {
                root = null;
            } else {
                if (p.color == BLACK) {
                    fixAfterDeletion(p);
                }
                if (p.parent != null) {
                    if (p == p.parent.left) {
                        p.parent.left = null;
                    } else {
                        if (p == p.parent.right) {
                            p.parent.right = null;
                        }
                    }
                    p.parent = null;
                }
            }
        }
    }

    private void fixAfterDeletion(final TreeSetEntry entry) {
        TreeSetEntry x = entry;
        while (x != root && colorOf(x) == BLACK) {
            if (x == leftOf(parentOf(x))) {
                TreeSetEntry sib = rightOf(parentOf(x));
                if (colorOf(sib) == RED) {
                    setColor(sib, BLACK);
                    setColor(parentOf(x), RED);
                    rotateLeft(parentOf(x));
                    sib = rightOf(parentOf(x));
                }

                if (colorOf(leftOf(sib)) == BLACK && colorOf(rightOf(sib)) == BLACK) {
                    setColor(sib, RED);
                    x = parentOf(x);
                } else {
                    if (colorOf(rightOf(sib)) == BLACK) {
                        setColor(leftOf(sib), BLACK);
                        setColor(sib, RED);
                        rotateRight(sib);
                        sib = rightOf(parentOf(x));
                    }
                    setColor(sib, colorOf(parentOf(x)));
                    setColor(parentOf(x), BLACK);
                    setColor(rightOf(sib), BLACK);
                    rotateLeft(parentOf(x));
                    x = root;
                }
            } else {
                TreeSetEntry sib = leftOf(parentOf(x));
                if (colorOf(sib) == RED) {
                    setColor(sib, BLACK);
                    setColor(parentOf(x), RED);
                    rotateRight(parentOf(x));
                    sib = leftOf(parentOf(x));
                }

                if (colorOf(rightOf(sib)) == BLACK && colorOf(leftOf(sib)) == BLACK) {
                    setColor(sib, RED);
                    x = parentOf(x);
                } else {
                    if (colorOf(leftOf(sib)) == BLACK) {
                        setColor(rightOf(sib), BLACK);
                        setColor(sib, RED);
                        rotateLeft(sib);
                        sib = leftOf(parentOf(x));
                    }

                    setColor(sib, colorOf(parentOf(x)));
                    setColor(parentOf(x), BLACK);
                    setColor(leftOf(sib), BLACK);
                    rotateRight(parentOf(x));
                    x = root;
                }
            }
        }

        setColor(x, BLACK);
    }

    private void decrementSize() {
        modCount++;
        size--;
    }

    private TreeSetEntry successor(TreeSetEntry t) {
        if (t == null) {
            return null;
        } else {
            if (t.right != null) {
                TreeSetEntry p = t.right;
                while (p.left != null) {
                    p = p.left;
                }
                return p;
            } else {
                TreeSetEntry p = t.parent;
                TreeSetEntry ch = t;
                while (p != null && ch == p.right) {
                    ch = p;
                    p = p.parent;
                }
                return p;
            }
        }
    }

}
